"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import glob
import hashlib
import imp
import inspect
import logging
import os
import sys
import types
from importlib.abc import FileLoader, Finder
from typing import Dict, Iterable, List, Optional, Set, Tuple

import pkg_resources

from inmanta import const

VERSION_FILE = "version"
MODULE_DIR = "modules"

LOGGER = logging.getLogger(__name__)


class SourceNotFoundException(Exception):
    """ This exception is raised when the source of the provided type is not found
    """


class SourceInfo(object):
    """ This class is used to store information related to source code information
    """

    def __init__(self, path: str, module_name: str) -> None:
        """
            :param path: The path of the source code file
            :param path: The name of the inmanta module
        """
        self.path = path
        self._hash: Optional[str] = None
        self._content: Optional[str] = None
        self._requires: Optional[List[str]] = None
        self.module_name = module_name

    @property
    def hash(self) -> str:
        """ Get the sha1 hash of the file
        """
        if self._hash is None:
            sha1sum = hashlib.new("sha1")
            sha1sum.update(self.content.encode("utf-8"))
            self._hash = sha1sum.hexdigest()

        return self._hash

    @property
    def content(self) -> str:
        """ Get the content of the file
        """
        if self._content is None:
            with open(self.path, "r", encoding="utf-8") as fd:
                self._content = fd.read()
        return self._content

    def _get_module_name(self) -> str:
        """Get the name of the inmanta module, derived from the python module name
        """
        module_parts = self.module_name.split(".")
        if module_parts[0] != const.PLUGINS_PACKAGE:
            raise Exception(
                "All instances from which the source is loaded, should be defined in the inmanta plugins package. "
                "%s does not match" % self.module_name
            )

        return module_parts[1]

    @property
    def requires(self) -> List[str]:
        """ List of python requirements associated with this source file
        """
        from inmanta.module import Project

        if self._requires is None:
            self._requires = Project.get().modules[self._get_module_name()].get_python_requirements_as_list()
        return self._requires


class CodeManager(object):
    """ This class is responsible for loading and packaging source code for types (resources, handlers, ...) that need to be
    available in a remote process (e.g. agent).
    """

    def __init__(self) -> None:
        self.__type_file: Dict[str, Set[str]] = {}
        self.__file_info: Dict[str, SourceInfo] = {}

    def register_code(self, type_name: str, instance: object) -> None:
        """ Register the given type_object under the type_name and register the source associated with this type object.

        :param type_name: The inmanta type name for which the source of type_object will be registered. For example std::File
        :param instance: An instance for which the code needs to be registered.
        """
        file_name = self.get_object_source(instance)
        if file_name is None:
            raise SourceNotFoundException("Unable to locate source code of instance %s for entity %s" % (inspect, type_name))

        if type_name not in self.__type_file:
            self.__type_file[type_name] = set()

        if file_name in self.__type_file[type_name]:
            return

        self.__type_file[type_name].add(file_name)

        if file_name not in self.__file_info:
            self.__file_info[file_name] = SourceInfo(file_name, instance.__module__)

    def get_object_source(self, instance: object) -> Optional[str]:
        """ Get the path of the source file in which type_object is defined
        """
        try:
            return inspect.getsourcefile(instance)
        except TypeError:
            return None

    def get_file_hashes(self) -> Iterable[str]:
        """ Return the hashes of all source files
        """
        return (info.hash for info in self.__file_info.values())

    def get_file_content(self, hash: str) -> str:
        """ Get the file content for the given hash
        """
        for info in self.__file_info.values():
            if info.hash == hash:
                return info.content

        raise KeyError("No file found with this hash")

    def get_types(self) -> Iterable[Tuple[str, List[SourceInfo]]]:
        """ Get a list of all registered types
        """
        return ((type_name, [self.__file_info[path] for path in files]) for type_name, files in self.__type_file.items())


class CodeLoader(object):
    """
        Class responsible for managing code loaded from modules received from the compiler

        :param code_dir: The directory where the code is stored
    """

    def __init__(self, code_dir: str) -> None:
        self.__code_dir = code_dir
        self.__modules: Dict[str, Tuple[str, types.ModuleType]] = {}  # A map with all modules we loaded, and its hv

        self.__check_dir()
        self.load_modules()

    def load_modules(self) -> None:
        """
            Load all existing modules
        """
        mod_dir = os.path.join(self.__code_dir, MODULE_DIR)
        pkg_resources.working_set = pkg_resources.WorkingSet._build_master()

        for py in glob.glob(os.path.join(mod_dir, "*.py")):
            if mod_dir in py:
                mod_name = py[len(mod_dir) + 1 : -3]
            else:
                mod_name = py[:-3]

            with open(py, "r", encoding="utf-8") as fd:
                source_code = fd.read().encode("utf-8")

            sha1sum = hashlib.new("sha1")
            sha1sum.update(source_code)

            hv = sha1sum.hexdigest()

            self._load_module(mod_name, py, hv)

    def __check_dir(self) -> None:
        """
            Check if the code directory
        """
        # check for the code dir
        if not os.path.exists(self.__code_dir):
            os.makedirs(self.__code_dir, exist_ok=True)

        # check for modules subdir
        if not os.path.exists(os.path.join(self.__code_dir, MODULE_DIR)):
            os.makedirs(os.path.join(self.__code_dir, MODULE_DIR), exist_ok=True)

    def _load_module(self, mod_name: str, python_file: str, hv: str) -> None:
        """
            Load or reload a module
        """
        try:
            mod = imp.load_source(mod_name, python_file)
            self.__modules[mod_name] = (hv, mod)
            LOGGER.info("Loaded module %s" % mod_name)
        except ImportError:
            LOGGER.exception("Unable to load module %s" % mod_name)

    def deploy_version(self, hash_value: str, module_name: str, module_source: str) -> None:
        """ Deploy a new version of the modules
        """
        # if the module is new, or update
        if module_name not in self.__modules or hash_value != self.__modules[module_name][0]:
            LOGGER.info("Deploying code (hv=%s, module=%s)", hash_value, module_name)
            # write the new source
            source_file = os.path.join(self.__code_dir, MODULE_DIR, module_name + ".py")

            fd = open(source_file, "w+", encoding="utf-8")
            fd.write(module_source)
            fd.close()

            # (re)load the new source
            self._load_module(module_name, source_file, hash_value)

        pkg_resources.working_set = pkg_resources.WorkingSet._build_master()


class PluginModuleLoader(FileLoader):
    """
        A custom module loader which imports the modules in the inmanta_plugins package.
    """

    def __init__(self, modulepaths: List[str], fullname: str) -> None:
        self._modulepaths = modulepaths
        path_to_module = self._get_path_to_module(fullname)
        super(PluginModuleLoader, self).__init__(fullname, path_to_module)

    def get_source(self, fullname: str) -> bytes:
        # No __init__.py exists for top level package
        if self._loading_top_level_package():
            return "".encode("utf-8")
        with open(self.path, "r", encoding="utf-8") as fd:
            return fd.read().encode("utf-8")

    def is_package(self, fullname: str) -> bool:
        if self._loading_top_level_package():
            return True
        return os.path.basename(self.path) == "__init__.py"

    def _get_path_to_module(self, fullname: str):
        module_parts = fullname.split(".")[1:]
        # No __init__.py exists for top level package
        if len(module_parts) == 0:
            return ""
        module_parts.insert(1, "plugins")
        for module_path in self._modulepaths:
            path_to_module = os.path.join(module_path, *module_parts)
            if os.path.exists(f"{path_to_module}.py"):
                return f"{path_to_module}.py"
            if os.path.isdir(path_to_module):
                path_to_module = os.path.join(path_to_module, "__init__.py")
                if os.path.exists(path_to_module):
                    return path_to_module

        raise ImportError(f"Cannot find module {fullname} in {self._modulepaths}")

    def _loading_top_level_package(self):
        return self.path == ""


class PluginModuleFinder(Finder):
    """
        Custom module finder which handles all the imports for the package inmanta_plugins.
    """

    def __init__(self, modulepaths: List[str]) -> None:
        self._modulepaths = modulepaths

    def add_module_paths(self, paths: List[str]) -> None:
        for p in paths:
            if p not in self._modulepaths:
                self._modulepaths.append(p)

    def find_module(self, fullname: str, path: Optional[str] = None) -> Optional[PluginModuleLoader]:
        if fullname == const.PLUGINS_PACKAGE or fullname.startswith(f"{const.PLUGINS_PACKAGE}."):
            LOGGER.debug("Loading module: %s", fullname)
            return PluginModuleLoader(self._modulepaths, fullname)
        return None


def configure_module_finder(modulepaths: List[str]) -> None:
    """
        Setup a custom module loader to handle imports in .py files of the modules.
    """
    for finder in sys.meta_path:
        # PluginModuleFinder already present in sys.meta_path.
        if isinstance(finder, PluginModuleFinder):
            finder.add_module_paths(modulepaths)
            return

    # PluginModuleFinder not yet present in sys.meta_path.
    module_finder = PluginModuleFinder(modulepaths)
    sys.meta_path.insert(0, module_finder)


def unload_inmanta_plugins():
    """
        Unload the inmanta_plugins package.
    """
    loaded_modules = sys.modules.keys()
    modules_to_unload = [k for k in loaded_modules if k == const.PLUGINS_PACKAGE or k.startswith(f"{const.PLUGINS_PACKAGE}.")]
    for k in modules_to_unload:
        del sys.modules[k]



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import inspect
import logging
import socket
import uuid
from asyncio import CancelledError, sleep
from collections import defaultdict
from enum import Enum
from typing import Any, Callable, Coroutine, Dict, Generator, List, Optional, Set, Tuple, Union  # noqa: F401
from urllib import parse

from tornado import ioloop

from inmanta import config as inmanta_config
from inmanta import util
from inmanta.protocol.common import UrlMethod
from inmanta.util import TaskHandler

from . import common
from .rest import client

LOGGER: logging.Logger = logging.getLogger(__name__)


class CallTarget(object):
    """
        A baseclass for all classes that are target for protocol calls / methods
    """

    def _get_endpoint_metadata(self) -> Dict[str, List[Tuple[str, Callable]]]:
        total_dict = {
            method_name: method
            for method_name, method in inspect.getmembers(self)
            if callable(method) and method_name[0] != "_"
        }

        methods: Dict[str, List[Tuple[str, Callable]]] = defaultdict(list)
        for name, attr in total_dict.items():
            if hasattr(attr, "__protocol_method__"):
                methods[attr.__protocol_method__.__name__].append((name, attr))

        return methods

    def get_op_mapping(self) -> Dict[str, Dict[str, UrlMethod]]:
        """
            Build a mapping between urls, ops and methods
        """
        url_map: Dict[str, Dict[str, UrlMethod]] = defaultdict(dict)

        # Loop over all methods in this class that have a handler annotation. The handler annotation refers to a method
        # definition. This method definition defines how the handler is invoked.
        for method, handler_list in self._get_endpoint_metadata().items():
            for method_handlers in handler_list:
                # Go over all method annotation on the method associated with the handler
                for properties in common.MethodProperties.methods[method]:
                    url = properties.get_listen_url()

                    # Associate the method with the handler if:
                    # - the handler does not specific a method version
                    # - the handler specifies a method version and the method version matches the method properties
                    if method_handlers[1].__api_version__ is None or (
                        method_handlers[1].__api_version__ is not None
                        and properties.api_version == method_handlers[1].__api_version__
                    ):
                        # there can only be one
                        if url in url_map and properties.operation in url_map[url]:
                            raise Exception(f"A handler is already registered for {properties.operation} {url}. ")

                        url_map[url][properties.operation] = UrlMethod(properties, self, method_handlers[1], method_handlers[0])
        return url_map


class Endpoint(TaskHandler):
    """
        An end-point in the rpc framework
    """

    def __init__(self, name: str):
        super(Endpoint, self).__init__()
        self._name: str = name
        self._node_name: str = inmanta_config.nodename.get()
        self._end_point_names: Set[str] = set()
        self._targets: List[CallTarget] = []

    def add_call_target(self, target: CallTarget) -> None:
        self._targets.append(target)

    @property
    def call_targets(self) -> List[CallTarget]:
        return self._targets

    def get_end_point_names(self) -> Set[str]:
        return self._end_point_names

    async def add_end_point_name(self, name: str) -> None:
        """
            Add an additional name to this endpoint to which it reacts and sends out in heartbeats
        """
        LOGGER.debug("Adding '%s' as endpoint", name)
        self._end_point_names.add(name)

    async def remove_end_point_name(self, name: str) -> None:
        LOGGER.debug("Removing '%s' as endpoint", name)
        self._end_point_names.discard(name)

    def clear_end_points(self) -> None:
        """
            Clear all endpoints
        """
        self._end_point_names = set()

    name = property(lambda self: self._name)
    end_point_names = property(get_end_point_names)

    def _get_hostname(self) -> str:
        """
            Determine the hostname of this machine
        """
        return socket.gethostname()

    def get_node_name(self) -> str:
        return self._node_name

    node_name = property(get_node_name)

    async def stop(self) -> None:
        """ Stop this endpoint
        """
        await super(Endpoint, self).stop()


class SessionEndpoint(Endpoint, CallTarget):
    """
        An endpoint for clients that make calls to a server and that receive calls back from the server using long-poll
    """

    _client: "SessionClient"

    def __init__(self, name: str, timeout: int = 120, reconnect_delay: int = 5):
        super().__init__(name)
        self._transport = client.RESTClient
        self._sched = util.Scheduler("session endpoint")

        self._env_id: Optional[uuid.UUID] = None

        self.sessionid: uuid.UUID = uuid.uuid1()
        self.running: bool = True
        self.server_timeout = timeout
        self.reconnect_delay = reconnect_delay
        self.add_call_target(self)

    def get_environment(self) -> Optional[uuid.UUID]:
        return self._env_id

    @property
    def environment(self) -> Optional[uuid.UUID]:
        return self._env_id

    def set_environment(self, environment_id: uuid.UUID) -> None:
        """
            Set the environment of this agent
        """
        if isinstance(environment_id, str):
            self._env_id = uuid.UUID(environment_id)
        else:
            self._env_id = environment_id

    async def start_connected(self):
        """
            This method is called after starting the client transport, but before sending the first heartbeat.
        """
        pass

    async def start(self) -> None:
        """
            Connect to the server and use a heartbeat and long-poll for two-way communication
        """
        assert self._env_id is not None
        LOGGER.log(3, "Starting agent for %s", str(self.sessionid))
        self._client = SessionClient(self.name, self.sessionid, timeout=self.server_timeout)
        await self.start_connected()
        self.add_background_task(self.perform_heartbeat())

    async def stop(self) -> None:
        self._sched.stop()
        await super(SessionEndpoint, self).stop()

    async def on_reconnect(self) -> None:
        """
        Called when a connection becomes active. i.e. when a first heartbeat is received after startup or
        a first hearbeat after an :py:`on_disconnect`
        """
        pass

    async def on_disconnect(self) -> None:
        """
        Called when the connection is lost unexpectedly (not on shutdown)
        """
        pass

    async def perform_heartbeat(self) -> None:
        """
            Start a continuous heartbeat call
        """
        if self._client is None:
            raise Exception("AgentEndpoint not started")

        connected: bool = False
        try:
            while True:
                LOGGER.log(3, "sending heartbeat for %s", str(self.sessionid))
                result = await self._client.heartbeat(
                    sid=str(self.sessionid),
                    tid=str(self._env_id),
                    endpoint_names=list(self.end_point_names),
                    nodename=self.node_name,
                    no_hang=not connected,
                )
                LOGGER.log(3, "returned heartbeat for %s", str(self.sessionid))
                if result.code == 200:
                    if not connected:
                        connected = True
                        self.add_background_task(self.on_reconnect())
                    if result.result is not None:
                        if "method_calls" in result.result:
                            method_calls: List[common.Request] = [
                                common.Request.from_dict(req) for req in result.result["method_calls"]
                            ]
                            # FIXME: reuse transport?
                            transport = self._transport(self)

                            for method_call in method_calls:
                                self.add_background_task(self.dispatch_method(transport, method_call))
                else:
                    LOGGER.warning(
                        "Heartbeat failed with status %d and message: %s, going to sleep for %d s",
                        result.code,
                        result.result,
                        self.reconnect_delay,
                    )
                    connected = False
                    await self.on_disconnect()
                    await sleep(self.reconnect_delay)
        except CancelledError:
            pass

    async def dispatch_method(self, transport: client.RESTClient, method_call: common.Request) -> None:
        if self._client is None:
            raise Exception("AgentEndpoint not started")

        LOGGER.debug("Received call through heartbeat: %s %s %s", method_call.reply_id, method_call.method, method_call.url)
        kwargs, config = transport.match_call(method_call.url, method_call.method)

        if config is None:
            msg = "An error occurred during heartbeat method call (%s %s %s): %s" % (
                method_call.reply_id,
                method_call.method,
                method_call.url,
                "No such method",
            )
            LOGGER.error(msg)
            self.add_background_task(
                self._client.heartbeat_reply(self.sessionid, method_call.reply_id, {"result": msg, "code": 500})
            )

        body = method_call.body or {}
        query_string = parse.urlparse(method_call.url).query
        for key, value in parse.parse_qs(query_string, keep_blank_values=True):
            if len(value) == 1:
                body[key] = value[0].decode("latin-1")
            else:
                body[key] = [v.decode("latin-1") for v in value]

        # FIXME: why create a new transport instance on each call? keep-alive?
        response: common.Response = await transport._execute_call(kwargs, method_call.method, config, body, method_call.headers)

        if response.status_code == 500:
            msg = ""
            if response.body is not None and "message" in response.body:
                msg = response.body["message"]
            LOGGER.error(
                "An error occurred during heartbeat method call (%s %s %s): %s",
                method_call.reply_id,
                method_call.method,
                method_call.url,
                msg,
            )

        if self._client is None:
            raise Exception("AgentEndpoint not started")

        await self._client.heartbeat_reply(
            self.sessionid, method_call.reply_id, {"result": response.body, "code": response.status_code}
        )


class VersionMatch(str, Enum):
    lowest = "lowest"
    """ Select the lowest available version of the method
    """
    highest = "highest"
    """ Select the highest available version of the method
    """
    exact = "exact"
    """ Select the exact version of the method
    """


class Client(Endpoint):
    """
        A client that communicates with end-point based on its configuration
    """

    def __init__(
        self,
        name: str,
        timeout: int = 120,
        version_match: VersionMatch = VersionMatch.lowest,
        exact_version: int = 0,
        with_rest_client: bool = True,
    ) -> None:
        super().__init__(name)
        assert isinstance(timeout, int), "Timeout needs to be an integer value."
        LOGGER.debug("Start transport for client %s", self.name)
        if with_rest_client:
            self._transport_instance = client.RESTClient(self, connection_timout=timeout)
        else:
            self._transport_instance = None
        self._version_match = version_match
        self._exact_version = exact_version

    async def _call(self, method_properties: common.MethodProperties, args: List, kwargs: Dict) -> common.Result:
        """
            Execute a call and return the result
        """
        result = await self._transport_instance.call(method_properties, args, kwargs)
        return result

    def _select_method(self, name) -> Optional[common.MethodProperties]:
        if name not in common.MethodProperties.methods:
            return None

        methods = common.MethodProperties.methods[name]

        if self._version_match is VersionMatch.lowest:
            return min(methods, key=lambda x: x.api_version)
        elif self._version_match is VersionMatch.highest:
            return max(methods, key=lambda x: x.api_version)
        elif self._version_match is VersionMatch.exact:
            for method in methods:
                if method.api_version == self._exact_version:
                    return method

        return None

    def __getattr__(self, name: str) -> Callable:
        """
            Return a function that will call self._call with the correct method properties associated
        """
        method = self._select_method(name)

        if method is None:
            raise AttributeError("Method with name %s is not defined for this client" % name)

        def wrap(*args: List, **kwargs: Dict) -> common.Result:
            method.function(*args, **kwargs)
            return self._call(method_properties=method, args=args, kwargs=kwargs)

        return wrap


class SyncClient(object):
    """
        A synchronous client that communicates with end-point based on its configuration
    """

    def __init__(self, name: str, timeout: int = 120) -> None:
        self.name = name
        self.timeout = timeout
        self._client = Client(self.name, self.timeout)

    def __getattr__(self, name: str) -> Callable:
        def async_call(*args: List, **kwargs: Dict) -> None:
            method = getattr(self._client, name)

            def method_call() -> None:
                return method(*args, **kwargs)

            try:
                return ioloop.IOLoop.current().run_sync(method_call, self.timeout)
            except TimeoutError:
                raise ConnectionRefusedError()

        return async_call


class SessionClient(Client):
    """
        A client that communicates with server endpoints over a session.
    """

    def __init__(self, name: str, sid: uuid.UUID, timeout: int = 120) -> None:
        super().__init__(name, timeout)
        self._sid = sid

    async def _call(self, method_properties: common.MethodProperties, args: List, kwargs: Dict) -> common.Result:
        """
            Execute the rpc call
        """
        if "sid" not in kwargs:
            kwargs["sid"] = self._sid

        result = await self._transport_instance.call(method_properties, args, kwargs)
        return result



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import pytest

import inmanta.compiler as compiler
from inmanta.ast import OptionalValueException
from inmanta.execute.util import NoneValue


def test_null(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
        entity A:
            string? a = null
        end
        implement A using std::none
        a = A()

    """
    )

    (_, scopes) = compiler.do_compile()
    root = scopes.get_child("__config__")
    a = root.lookup("a").get_value().get_attribute("a").get_value()
    assert isinstance(a, NoneValue)


def test_null_unset(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
        entity A:
            string? a
        end
        implement A using std::none
        a = A()

    """
    )

    (_, scopes) = compiler.do_compile()
    root = scopes.get_child("__config__")
    with pytest.raises(OptionalValueException):
        root.lookup("a").get_value().get_attribute("a").get_value()


def test_null_unset_hang(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
            entity A:
                string? a
            end
            implement A using std::none
            a = A()
            b = a.a
        """
    )
    with pytest.raises(OptionalValueException):
        (_, scopes) = compiler.do_compile()


def test_null_on_list(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
        entity A:
            string[]? a = null
        end
        implement A using std::none
        a = A()
    """
    )

    (_, scopes) = compiler.do_compile()
    root = scopes.get_child("__config__")
    a = root.lookup("a").get_value().get_attribute("a").get_value()
    assert isinstance(a, NoneValue)


def test_null_on_dict(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
        entity A:
            dict? a = null
        end
        implement A using std::none
        a = A()
    """
    )

    (_, scopes) = compiler.do_compile()
    root = scopes.get_child("__config__")
    a = root.lookup("a").get_value().get_attribute("a").get_value()
    assert isinstance(a, NoneValue)


def test_null_on_dict_err(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
        entity A:
            dict a = null
        end
        implement A using std::none
        a = A()
    """,
        'Syntax error: null can not be assigned to dict, did you mean "dict? a = null" ({dir}/main.cf:3:18)',
    )


def test_null_err(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
        entity A:
            string a = null
        end
        implement A using std::none
        a = A()

    """,
        "Invalid value 'null', expected String (reported in string a = null ({dir}/main.cf:3))",
    )


def test_null_on_list_err(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
        entity A:
            string[] a = null
        end
        implement A using std::none
        a = A()
    """,
        "Invalid value 'null', expected string[] (reported in string[] a = null ({dir}/main.cf:3))",
    )



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from typing import Dict

import pytest

import inmanta.compiler as compiler
from inmanta.ast import Namespace
from inmanta.execute.runtime import Instance, ResultVariable, Typeorvalue


@pytest.mark.parametrize("parents", [True, False])
def test_implement_parents(snippetcompiler, parents: bool):
    snippetcompiler.setup_for_snippet(
        """
entity Parent:
    number n
end

entity Child extends Parent:
    number m
end

implement Parent using p
implement Child using %s, generic


implementation p for Parent:
    self.n = 0
end

implementation c for Child:
    self.n = 1
end

implementation generic for Child:
    self.m = 0
end


x = Child()
        """
        % ("parents" if parents else "c"),
    )
    (_, scopes) = compiler.do_compile()
    root: Namespace = scopes.get_child("__config__")
    x: Typeorvalue = root.lookup("x")
    assert isinstance(x, ResultVariable)
    instance = x.get_value()
    assert isinstance(instance, Instance)

    expected_attrs: Dict[str, int] = {"n": 0 if parents else 1, "m": 0}
    for name, value in expected_attrs.items():
        attr: Typeorvalue = instance.lookup(name)
        assert isinstance(attr, ResultVariable)
        assert attr.get_value() == value


def test_implement_parents_conditional_error(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
entity Parent:
end

entity Child extends Parent:
end

implement Parent using std::none
implement Child using std::none
implement Child using parents when 1 == 1


x = Child()
        """,
        "Conditional implementation with parents not allowed (reported in Implement(Child) ({dir}/main.cf:10:11))",
    )



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import os

import pytest

from inmanta.command import CLIException
from inmanta.moduletool import ModuleTool
from moduletool.common import install_project
from test_app_cli import app


def test_freeze_basic(modules_dir, modules_repo):
    install_project(modules_dir, "projectA")
    modtool = ModuleTool()
    cmod = modtool.get_module("modC")
    assert cmod.get_freeze("modC", recursive=False, mode="==") == {"std": "== 3.2", "modE": "== 3.2", "modF": "== 3.2"}
    assert cmod.get_freeze("modC", recursive=True, mode="==") == {
        "std": "== 3.2",
        "modE": "== 3.2",
        "modF": "== 3.2",
        "modH": "== 3.2",
        "modJ": "== 3.2",
    }

    assert cmod.get_freeze("modC::a", recursive=False, mode="==") == {"std": "== 3.2", "modI": "== 3.2"}


def test_project_freeze_basic(modules_dir, modules_repo):
    install_project(modules_dir, "projectA")
    modtool = ModuleTool()
    proj = modtool.get_project()
    assert proj.get_freeze(recursive=False, mode="==") == {
        "std": "== 3.2",
        "modB": "== 3.2",
        "modC": "== 3.2",
        "modD": "== 3.2",
    }
    assert proj.get_freeze(recursive=True, mode="==") == {
        "std": "== 3.2",
        "modB": "== 3.2",
        "modC": "== 3.2",
        "modD": "== 3.2",
        "modE": "== 3.2",
        "modF": "== 3.2",
        "modG": "== 3.2",
        "modH": "== 3.2",
        "modJ": "== 3.2",
    }


def test_project_freeze_bad(modules_dir, modules_repo, capsys, caplog):
    coroot = install_project(modules_dir, "baddep", config=False)

    with pytest.raises(CLIException) as e:
        app(["project", "freeze"])

    assert e.value.exitcode == 1
    assert str(e.value) == "Could not load project"

    out, err = capsys.readouterr()

    assert len(err) == 0, err
    assert len(out) == 0, out
    assert "requirement mod2<2016 on module mod2 not fullfilled, now at version 2016.1" in caplog.text

    assert os.path.getsize(os.path.join(coroot, "project.yml")) != 0


def test_project_freeze(modules_dir, modules_repo, capsys):
    coroot = install_project(modules_dir, "projectA")

    app(["project", "freeze", "-o", "-"])

    out, err = capsys.readouterr()

    assert os.path.getsize(os.path.join(coroot, "project.yml")) != 0
    assert len(err) == 0, err
    assert (
        out
        == """name: projectA
license: Apache 2.0
version: 0.0.1
modulepath: libs
downloadpath: libs
repo: %s
requires:
- modB ~= 3.2
- modC ~= 3.2
- modD ~= 3.2
- std ~= 3.2
"""
        % modules_repo
    )


def test_project_freeze_disk(modules_dir, modules_repo, capsys):
    coroot = install_project(modules_dir, "projectA")

    app(["project", "freeze"])

    out, err = capsys.readouterr()

    assert os.path.getsize(os.path.join(coroot, "project.yml")) != 0
    assert len(err) == 0, err

    with open(os.path.join(coroot, "project.yml"), "r", encoding="utf-8") as fh:
        assert (
            fh.read()
            == """name: projectA
license: Apache 2.0
version: 0.0.1
modulepath: libs
downloadpath: libs
repo: %s
requires:
- modB ~= 3.2
- modC ~= 3.2
- modD ~= 3.2
- std ~= 3.2
"""
            % modules_repo
        )


def test_project_freeze_odd_opperator(modules_dir, modules_repo, capsys, caplog):
    coroot = install_project(modules_dir, "projectA")

    app(["project", "freeze", "-o", "-", "--operator", "xxx"])

    out, err = capsys.readouterr()

    assert os.path.getsize(os.path.join(coroot, "project.yml")) != 0
    assert len(err) == 0, err
    assert (
        out
        == """name: projectA
license: Apache 2.0
version: 0.0.1
modulepath: libs
downloadpath: libs
repo: %s
requires:
- modB xxx 3.2
- modC xxx 3.2
- modD xxx 3.2
- std xxx 3.2
"""
        % modules_repo
    )

    assert "Operator xxx is unknown, expecting one of ['==', '~=', '>=']" in caplog.text


def test_project_options_in_config(modules_dir, modules_repo, capsys):
    coroot = install_project(modules_dir, "projectA")
    with open("project.yml", "w", encoding="utf-8") as fh:
        fh.write(
            """name: projectA
license: Apache 2.0
version: 0.0.1
modulepath: libs
downloadpath: libs
repo: %s
freeze_recursive: true
freeze_operator: ==
"""
            % modules_repo
        )

    def verify():
        out, err = capsys.readouterr()

        assert os.path.getsize(os.path.join(coroot, "project.yml")) != 0
        assert len(err) == 0, err
        assert len(out) == 0, out

        with open("project.yml", "r", encoding="utf-8") as fh:
            assert fh.read() == (
                """name: projectA
license: Apache 2.0
version: 0.0.1
modulepath: libs
downloadpath: libs
repo: %s
freeze_recursive: true
freeze_operator: ==
requires:
- modB == 3.2
- modC == 3.2
- modD == 3.2
- modE == 3.2
- modF == 3.2
- modG == 3.2
- modH == 3.2
- modJ == 3.2
- std == 3.2
"""
                % modules_repo
            )

    app(["project", "freeze"])
    verify()
    app(["project", "freeze"])
    verify()


def test_module_freeze(modules_dir, modules_repo, capsys):
    coroot = install_project(modules_dir, "projectA")

    def verify():
        out, err = capsys.readouterr()

        assert os.path.getsize(os.path.join(coroot, "project.yml")) != 0
        assert len(err) == 0, err
        assert out == (
            """name: modC
license: Apache 2.0
version: '3.2'
requires:
- modE ~= 3.2
- modF ~= 3.2
- modI ~= 3.2
- std ~= 3.2
"""
        )

    app(["module", "-m", "modC", "freeze", "-o", "-"])
    verify()


def test_module_freeze_self_disk(modules_dir, modules_repo, capsys):
    coroot = install_project(modules_dir, "projectA")

    def verify():
        out, err = capsys.readouterr()

        assert len(err) == 0, err
        assert len(out) == 0, out

        modpath = os.path.join(coroot, "libs/modC/module.yml")
        assert os.path.getsize(os.path.join(coroot, "project.yml")) != 0
        assert os.path.getsize(modpath) != 0

        with open(modpath, "r", encoding="utf-8") as fh:
            outf = fh.read()
            assert outf == (
                """name: modC
license: Apache 2.0
version: '3.2'
requires:
- modE ~= 3.2
- modF ~= 3.2
- modI ~= 3.2
- std ~= 3.2
"""
            )

    modp = os.path.join(coroot, "libs/modC")
    app(["module", "install"])
    os.chdir(modp)
    os.curdir = modp
    app(["module", "freeze"])
    verify()



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from threading import Lock, Thread
from time import sleep

import pytest
from pytest import fixture

from inmanta.agent.cache import AgentCache
from inmanta.agent.handler import cache
from inmanta.resources import Id, Resource, resource


@fixture()
def my_resource():
    @resource("test::Resource", agent="agent", id_attribute="key")
    class MyResource(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged")


def test_base():
    cache = AgentCache()
    value = "test too"
    cache.cache_value("test", value)
    assert value == cache.find("test")


def test_timout():
    cache = AgentCache()
    value = "test too"
    cache.cache_value("test", value, timeout=0.1)
    cache.cache_value("test2", value)

    assert value == cache.find("test")
    sleep(0.2)
    try:
        assert value == cache.find("test")
        raise AssertionError("Should get exception")
    except KeyError:
        pass

    assert value == cache.find("test2")


def test_base_fail():
    cache = AgentCache()
    value = "test too"
    with pytest.raises(KeyError):
        assert value == cache.find("test")


def test_resource():
    cache = AgentCache()
    value = "test too"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    cache.cache_value("test", value, resource=resource)
    assert value == cache.find("test", resource=resource)


def test_resource_fail(my_resource):
    cache = AgentCache()
    value = "test too"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    cache.cache_value("test", value, resource=resource)

    with pytest.raises(KeyError):
        assert value == cache.find("test")


def test_version_closed():
    cache = AgentCache()
    value = "test too"
    version = 200
    with pytest.raises(Exception):
        cache.cache_value("test", value, version=version)
        assert value == cache.find("test", version=version)


def test_version():
    cache = AgentCache()
    value = "test too"
    version = 200
    cache.open_version(version)
    cache.cache_value("test", value, version=version)
    assert value == cache.find("test", version=version)


def test_version_close():
    cache = AgentCache()
    value = "test too"
    version = 200
    cache.open_version(version)
    cache.cache_value("test", value, version=version)
    cache.cache_value("test0", value, version=version)
    cache.cache_value("test4", value, version=version)
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    cache.cache_value("testx", value, resource=resource)
    assert value == cache.find("test", version=version)
    assert value == cache.find("testx", resource=resource)
    cache.close_version(version)
    assert value, cache.find("testx", resource=resource)
    with pytest.raises(KeyError):
        assert value == cache.find("test", version=version)
        raise AssertionError("Should get exception")


def test_context_manager():
    cache = AgentCache()
    value = "test too"
    version = 200
    with cache.manager(version):
        cache.cache_value("test", value, version=version)
        cache.cache_value("test0", value, version=version)
        cache.cache_value("test4", value, version=version)
        resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
        cache.cache_value("testx", value, resource=resource)
        assert value == cache.find("test", version=version)
        assert value == cache.find("testx", resource=resource)

    assert value, cache.find("testx", resource=resource)
    with pytest.raises(KeyError):
        assert value == cache.find("test", version=version)


def test_multi_threaded():
    class Spy(object):
        def __init__(self):
            self.created = 0
            self.deleted = 0
            self.lock = Lock()

        def create(self):
            with self.lock:
                self.created += 1
            return self

        def delete(self):
            self.deleted += 1

    cache = AgentCache()
    version = 200

    cache.open_version(version)

    alpha = Spy()
    beta = Spy()
    alpha.lock.acquire()

    t1 = Thread(
        target=lambda: cache.get_or_else(
            "test", lambda version: alpha.create(), version=version, call_on_delete=lambda x: x.delete()
        )
    )
    t2 = Thread(
        target=lambda: cache.get_or_else(
            "test", lambda version: beta.create(), version=version, call_on_delete=lambda x: x.delete()
        )
    )

    t1.start()
    t2.start()

    alpha.lock.release()

    t1.join()
    t2.join()

    assert alpha.created + beta.created == 1
    assert alpha.deleted == 0
    assert beta.deleted == 0

    cache.close_version(version)

    assert alpha.created + beta.created == 1
    assert alpha.deleted == alpha.created
    assert beta.deleted == beta.created


def test_timout_and_version():
    cache = AgentCache()
    version = 200

    cache.open_version(version)
    value = "test too"
    cache.cache_value("test", value, version=version, timeout=0.1)
    cache.cache_value("testx", value)

    assert value == cache.find("test", version=version)
    assert value == cache.find("testx")

    sleep(0.2)
    assert value == cache.find("testx")

    cache.close_version(version)
    assert value == cache.find("testx")

    with pytest.raises(KeyError):
        cache.find("test", version=version)
    assert value == cache.find("testx")


def test_version_and_timout():
    cache = AgentCache()
    version = 200

    cache.open_version(version)
    value = "test too"
    cache.cache_value("test", value, version=version, timeout=0.1)
    cache.cache_value("testx", value)

    assert value == cache.find("test", version=version)
    assert value == cache.find("testx")

    cache.close_version(version)
    assert value == cache.find("testx")

    sleep(0.2)
    assert value == cache.find("testx")

    with pytest.raises(KeyError):
        cache.find("test", version=version)


def test_version_fail():
    cache = AgentCache()
    value = "test too"
    version = 200
    cache.open_version(version)
    cache.cache_value("test", value, version=version)

    with pytest.raises(KeyError):
        assert value == cache.find("test")


def test_resource_and_version():
    cache = AgentCache()
    value = "test too"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    version = 200
    cache.open_version(version)
    cache.cache_value("test", value, resource=resource, version=version)
    assert value == cache.find("test", resource=resource, version=version)


def test_get_or_else(my_resource):
    called = []

    def creator(param, resource, version):

        called.append("x")
        return param

    cache = AgentCache()
    value = "test too"
    value2 = "test too x"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    resourcev2 = Id("test::Resource", "test", "key", "test", 200).get_instance()
    assert 200 == resourcev2.id.version
    version = 200
    cache.open_version(version)
    assert value == cache.get_or_else("test", creator, resource=resource, version=version, param=value)
    assert value == cache.get_or_else("test", creator, resource=resource, version=version, param=value)
    assert len(called) == 1
    assert value == cache.get_or_else("test", creator, resource=resourcev2, version=version, param=value)
    assert len(called) == 1
    assert value2 == cache.get_or_else("test", creator, resource=resource, version=version, param=value2)


def test_get_or_else_none():
    called = []

    def creator(param, resource, version):
        called.append("x")
        return param

    class Sequencer(object):
        def __init__(self, sequence):
            self.seq = sequence
            self.count = 0

        def __call__(self, **kwargs):
            out = self.seq[self.count]
            self.count += 1
            return out

    cache = AgentCache()
    value = "test too"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    version = 100
    cache.open_version(version)
    assert None is cache.get_or_else("test", creator, resource=resource, version=version, cache_none=False, param=None)
    assert len(called) == 1
    assert None is cache.get_or_else("test", creator, resource=resource, version=version, cache_none=False, param=None)
    assert len(called) == 2
    assert value == cache.get_or_else("test", creator, resource=resource, version=version, cache_none=False, param=value)
    assert value == cache.get_or_else("test", creator, resource=resource, version=version, cache_none=False, param=value)
    assert len(called) == 3

    seq = Sequencer([None, None, "A"])
    assert None is cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
    assert seq.count == 1
    assert None is cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
    assert seq.count == 2
    assert "A" == cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
    assert seq.count == 3
    assert "A" == cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
    assert seq.count == 3
    assert "A" == cache.get_or_else("testx", seq, resource=resource, version=version, cache_none=False)
    assert seq.count == 3


def test_decorator():
    class Closeable:
        def __init__(self):
            self.closed = False

        def close(self):
            self.closed = True

    my_closable = Closeable()
    my_closable_2 = Closeable()

    xcache = AgentCache()

    class DT(object):
        def __init__(self, cache):
            self.cache = cache
            self.count = 0
            self.c2 = 0

        @cache()
        def test_method(self):
            self.count += 1
            return "x"

        @cache
        def test_method_2(self, version):
            self.count += 1
            return "x2"

        @cache(cacheNone=False)
        def test_method_3(self):
            self.c2 += 1
            if self.c2 < 2:
                return None
            else:
                return "X"

        @cache(call_on_delete=lambda x: x.close())
        def test_close(self, version):
            self.count += 1
            return my_closable

        @cache(call_on_delete=lambda x: x.close())
        def test_close_2(self):
            self.count += 1
            return my_closable_2

    test = DT(xcache)

    xcache.open_version(3)
    test.test_close(version=3)
    test.test_close_2()
    xcache.close()
    assert my_closable.closed
    assert my_closable_2.closed

    test.count = 0
    my_closable.closed = False

    assert "x" == test.test_method()
    assert "x" == test.test_method()
    assert "x" == test.test_method()
    assert 1 == test.count

    xcache.open_version(1)
    xcache.open_version(2)
    assert "x2" == test.test_method_2(version=1)
    assert "x2" == test.test_method_2(version=1)
    assert 2 == test.count
    assert "x2" == test.test_method_2(version=2)
    assert 3 == test.count
    xcache.close_version(1)
    xcache.open_version(1)
    assert "x2" == test.test_method_2(version=1)
    assert "x2" == test.test_method_2(version=1)
    assert 4 == test.count

    assert None is test.test_method_3()
    assert 1 == test.c2
    assert "X" == test.test_method_3()
    assert 2 == test.c2
    assert "X" == test.test_method_3()
    assert 2 == test.c2

    test.count = 0
    xcache.open_version(3)
    test.test_close(version=3)
    assert test.count == 1
    test.test_close(version=3)
    assert test.count == 1
    assert not my_closable.closed
    xcache.close_version(3)
    assert my_closable.closed



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import re

import pytest
import tornado
from pyformance import timer
from tornado.httpserver import HTTPServer
from tornado.web import url

from inmanta.reporter import AsyncReporter, InfluxReporter


class QueryMockHandler(tornado.web.RequestHandler):
    def initialize(self, parent):
        self.parent = parent

    def get(self, *args, **kwargs):
        self.parent.querycount += 1
        try:
            assert self.request.query_arguments["q"] == [b"CREATE DATABASE metrics"]
        except Exception as e:
            # carry over  failures
            self.parent.failure = e


influxlineprotocol = re.compile(r"\w+(,\w+=\w+)* (\w+=[\d.e+-]*)(,\w+=[\d.e+-]*)* \d+")


class WriteMockHandler(tornado.web.RequestHandler):
    def initialize(self, parent):
        self.parent = parent

    def post(self, *args, **kwargs):
        self.parent.writecount += 1
        try:
            for line in self.request.body.decode().split("\n"):
                print(line)
                assert influxlineprotocol.match(line)
                self.parent.lines.append(line)
        except Exception as e:
            # carry over  failures
            self.parent.failure = e


class InfluxdbMock(object):
    def __init__(self, socket):
        self.querycount = 0
        self.writecount = 0
        self.failure = None
        self.lines = []

        self.app = tornado.web.Application(
            [
                url(r"/query", QueryMockHandler, kwargs={"parent": self}),
                url(r"/write", WriteMockHandler, kwargs={"parent": self}),
            ]
        )

        self.server = HTTPServer(self.app)
        self.server.add_sockets([socket])
        _addr, port = socket.getsockname()
        self.port = port


class MockReporter(AsyncReporter):
    def __init__(self, interval):
        super().__init__(None, interval)
        self.waiter = asyncio.locks.Semaphore(0)
        self.in_count = 0
        self.count = 0

    async def report_now(self, registry=None, timestamp=None) -> None:
        self.in_count += 1
        await self.waiter.acquire()
        self.count += 1


@pytest.fixture
def influxdb(event_loop, free_socket):
    ifl = InfluxdbMock(free_socket())
    yield ifl
    ifl.server.stop()


@pytest.mark.asyncio
async def test_influxdb(influxdb):
    rep = InfluxReporter(port=influxdb.port, tags={"mark": "X"}, autocreate_database=True)
    with timer("test").time():
        pass

    with timer("test2").time():
        pass

    await rep.report_now()

    assert influxdb.querycount == 1
    assert influxdb.writecount == 1

    if influxdb.failure:
        raise influxdb.failure

    for line in influxdb.lines:
        assert "mark=X" in line


@pytest.mark.asyncio
async def test_timing():
    # Attempt to deploy every 0.01 seconds,
    # Deploy hangs on a semaphore, so test case can control progress
    mr = MockReporter(0.01)
    mr.start()

    # no deploy done
    assert mr.count == 0
    # wait for deploy
    await asyncio.sleep(0.01)
    # one is waiting
    assert mr.in_count == 1
    # release
    mr.waiter.release()
    # wait 0 to allow reporter to make progress
    await asyncio.sleep(0.0)
    # one deploy done
    assert mr.count == 1

    # allow to run free to test timing
    for i in range(5):
        mr.waiter.release()

    # wait 0 to allow reporter to make progress
    await asyncio.sleep(0)
    # how many are we now?
    base = mr.count
    # should be sufficiently far of the lock
    # otherwise test always succeeds
    assert base < 4
    # wait for one more
    await asyncio.sleep(0.01)
    assert mr.count == base + 1
    mr.stop()



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import pytest

from inmanta.export import DependencyCycleException
from utils import assert_graph


def test_abstract_requires(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
host = std::Host(name="host", os=std::unix)

entity A:
    string name
end

implementation a for A:
    one = std::ConfigFile(path="{{self.name}}1", host=host, content="")
    two = std::ConfigFile(path="{{self.name}}2", host=host, content="")
    two.requires = one
end

implement A using a

pre = std::ConfigFile(path="host0", host=host, content="")
post = std::ConfigFile(path="hosts4", host=host, content="")

inter = A(name = "inter")
"""
    )

    v, resources = snippetcompiler.do_export()
    assert_graph(resources, """inter2: inter1""")


def test_abstract_requires_3(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
host = std::Host(name="host", os=std::unix)

entity A:
    string name
end

implementation a for A:
    one = std::ConfigFile(path="{{self.name}}1", host=host, content="")
    two = std::ConfigFile(path="{{self.name}}2", host=host, content="")
    two.requires = one
    one.requires = self.requires
    two.provides = self.provides
end

implement A using a

pre = std::ConfigFile(path="pre", host=host, content="")
post = std::ConfigFile(path="post", host=host, content="")

inter = A(name = "inter")
inter.requires = pre
post.requires = inter
"""
    )

    v, resources = snippetcompiler.do_export()
    assert_graph(
        resources,
        """post: inter2
                                  inter2: inter1
                                  inter1: pre""",
    )


def test_abstract_requires_2(snippetcompiler, caplog):
    snippetcompiler.setup_for_snippet(
        """
host = std::Host(name="host", os=std::unix)

entity A:
string name
end

implementation a for A:
one = std::ConfigFile(path="{{self.name}}1", host=host, content="")
two = std::ConfigFile(path="{{self.name}}2", host=host, content="")
two.requires = one
end

implement A using a

pre = std::ConfigFile(path="host0", host=host, content="")
post = std::ConfigFile(path="hosts4", host=host, content="")

inter = A(name = "inter")
inter.requires = pre
post.requires = inter
"""
    )

    snippetcompiler.do_export()
    warning = [
        x
        for x in caplog.records
        if x.msg == "The resource %s had requirements before flattening, but not after flattening."
        " Initial set was %s. Perhaps provides relation is not wired through correctly?"
    ]
    assert len(warning) == 1


def test_issue_220_dep_loops(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import std

host = std::Host(name="Test", os=std::unix)
f1 = std::ConfigFile(host=host, path="/f1", content="")
f2 = std::ConfigFile(host=host, path="/f2", content="")
f3 = std::ConfigFile(host=host, path="/f3", content="")
f4 = std::ConfigFile(host=host, path="/f4", content="")
f1.requires = f2
f2.requires = f3
f3.requires = f1
f4.requires = f1
"""
    )
    with pytest.raises(DependencyCycleException) as e:
        snippetcompiler.do_export()

    cyclenames = [r.id.resource_str() for r in e.value.cycle]
    assert set(cyclenames) == set(["std::File[Test,path=/f3]", "std::File[Test,path=/f2]", "std::File[Test,path=/f1]"])



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import typing

import pytest

import inmanta.ast.type as inmanta_type
import inmanta.compiler as compiler
from inmanta.ast import AttributeException, Namespace, TypingException
from inmanta.ast.attribute import Attribute
from inmanta.ast.entity import Entity
from inmanta.ast.type import Bool, Integer, Number, String


def test_lnr_on_double_is_defined(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    string? two
end

Test.one [0:1] -- Test

implement Test using std::none when self.one.two is defined

a = Test(two="b")
a.one = a
"""
    )
    compiler.do_compile()


def test_double_define(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    string test
    string? test
    bool test
end
"""
    )
    with pytest.raises(TypingException):
        compiler.do_compile()


def test_536_number_cast(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Network:
    number segmentation_id
end
implement Network using std::none
net1 = Network(segmentation_id="10")
"""
    )
    with pytest.raises(AttributeException):
        compiler.do_compile()


def test_int_type(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    int i = 0
end

implement Test using std::none

Test(i = 42)
Test(i = -42)
Test()
        """,
    )
    compiler.do_compile()


def test_int_type_invalid(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
entity Test:
    int i = 0.0
end
        """,
        "Invalid value '0.0', expected int (reported in int i = 0.0 ({dir}/main.cf:3))",
    )


def test_cast_to_number(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
x = number("42")
x = 42
y = number("42.0")
y = 42.0
z = number(true)
z = 1
u = number(false)
u = 0
        """,
    )
    (_, scopes) = compiler.do_compile()
    root: Namespace = scopes.get_child("__config__")
    x = root.lookup("x").get_value()
    y = root.lookup("y").get_value()
    z = root.lookup("z").get_value()
    u = root.lookup("u").get_value()
    assert Number().validate(x)
    assert Number().validate(y)
    assert Number().validate(z)
    assert Number().validate(u)


def test_cast_to_int(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
x = int("42")
x = 42
y = int(true)
y = 1
z = int(false)
z = 0
        """,
    )
    (_, scopes) = compiler.do_compile()
    root: Namespace = scopes.get_child("__config__")
    x = root.lookup("x").get_value()
    y = root.lookup("y").get_value()
    z = root.lookup("z").get_value()
    assert Integer().validate(x)
    assert Integer().validate(y)
    assert Integer().validate(z)


def test_cast_to_string(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
x = string(42)
x = "42"
y = string(42.0)
y = "42.0"
z = string(true)
z = "true"
        """,
    )
    (_, scopes) = compiler.do_compile()
    root: Namespace = scopes.get_child("__config__")
    x = root.lookup("x").get_value()
    y = root.lookup("y").get_value()
    z = root.lookup("z").get_value()
    assert String().validate(x)
    assert String().validate(y)
    assert String().validate(z)


def test_cast_to_bool(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
x = bool(42)
x = true
y = bool(42.0)
y = true
z = bool(1)
z = true
u = bool(0)
u = false
v = bool("false")
v = true
w = bool("")
w = false
        """,
    )
    (_, scopes) = compiler.do_compile()
    root: Namespace = scopes.get_child("__config__")
    x = root.lookup("x").get_value()
    y = root.lookup("y").get_value()
    z = root.lookup("z").get_value()
    u = root.lookup("u").get_value()
    v = root.lookup("v").get_value()
    w = root.lookup("w").get_value()
    assert Bool().validate(x)
    assert Bool().validate(y)
    assert Bool().validate(z)
    assert Bool().validate(u)
    assert Bool().validate(v)
    assert Bool().validate(w)


def test_cast_exception_kwargs(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
bool(value = 1)
        """,
        "Only positional arguments allowed in type cast (reported in bool(value=1) ({dir}/main.cf:2))",
    )


def test_cast_exception_wrapped_kwargs(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
dct = {"value" : 1}
bool(**dct)
        """,
        "Only positional arguments allowed in type cast (reported in bool(**dct) ({dir}/main.cf:3))",
    )


def test_cast_exception_too_many_args(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
bool(0, 1)
        """,
        "Illegal arguments 0,1: type cast expects exactly 1 argument (reported in bool(0,1) ({dir}/main.cf:2))",
    )


def test_cast_exception_non_primitive(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
list("[]")
        """,
        "Can not call 'list', can only call plugin or primitive type cast (reported in list('[]') ({dir}/main.cf:2))",
    )


def test_cast_exception_type_error(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
number(null)
        """,
        "Failed to cast 'null' to number (reported in number(null) ({dir}/main.cf:2))",
    )


def test_cast_exception_value_error(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
number("Hello World!")
        """,
        "Failed to cast 'Hello World!' to number (reported in number('Hello World!') ({dir}/main.cf:2))",
    )


def test_cast_exception_int_value_error(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
int("0.0")
        """,
        "Failed to cast '0.0' to int (reported in int('0.0') ({dir}/main.cf:2))",
    )


@pytest.mark.parametrize(
    "parent_modifier,child_modifier", [("[]", ""), ("", "[]"), ("?", ""), ("", "?")],
)
def test_2132_inheritance_type_override(snippetcompiler, parent_modifier: str, child_modifier: str) -> None:
    snippetcompiler.setup_for_error(
        f"""
entity Parent:
    number{parent_modifier} n
end

entity Child extends Parent:
    number{child_modifier} n
end
        """,
        "Incompatible attributes (original at ({dir}/main.cf:7)) (duplicate at ({dir}/main.cf:3))",
    )


def test_attribute_type(snippetcompiler) -> None:
    snippetcompiler.setup_for_snippet(
        """
entity A:
    number n
    number[] ns
    number? maybe_n
    number[]? maybe_ns
    list lst
end
        """,
    )
    types: typing.Dict[str, inmanta_type.Type]
    (types, scopes) = compiler.do_compile()
    assert "__config__::A" in types
    entity = types["__config__::A"]
    assert isinstance(entity, Entity)
    attrs: typing.Dict[str, Attribute] = entity.get_attributes()

    assert "n" in attrs
    assert isinstance(attrs["n"].type, Number)

    assert "ns" in attrs
    ns_type: inmanta_type.Type = attrs["ns"].type
    assert isinstance(ns_type, inmanta_type.TypedList)
    assert isinstance(ns_type.element_type, Number)

    assert "maybe_n" in attrs
    maybe_n_type: inmanta_type.Type = attrs["maybe_n"].type
    assert isinstance(maybe_n_type, inmanta_type.NullableType)
    assert isinstance(maybe_n_type.element_type, Number)

    assert "maybe_ns" in attrs
    maybe_ns_type: inmanta_type.Type = attrs["maybe_ns"].type
    assert isinstance(maybe_ns_type, inmanta_type.NullableType)
    maybe_ns_element_type: inmanta_type.Type = maybe_ns_type.element_type
    assert isinstance(maybe_ns_element_type, inmanta_type.TypedList)
    assert isinstance(maybe_ns_element_type.element_type, Number)

    assert "lst" in attrs
    assert isinstance(attrs["lst"].type, inmanta_type.List)


@pytest.mark.parametrize("base_type", inmanta_type.TYPES.values())
def test_types_base_type(snippetcompiler, base_type: inmanta_type.Type) -> None:
    modified_types: typing.List[inmanta_type.Type] = [
        base_type,
        inmanta_type.TypedList(base_type),
        inmanta_type.NullableType(base_type),
        inmanta_type.NullableType(inmanta_type.TypedList(base_type)),
    ]
    for t in modified_types:
        # verify get_base_type returns the inmanta base type
        assert t.get_base_type().type_string() == base_type.type_string()
        # verify with_base_type is round-trip compatible
        assert t.with_base_type(base_type).type_string() == t.type_string()



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import os
import subprocess

import pytest

from inmanta import module
from inmanta.ast import CompilerException, ModuleNotFoundException
from inmanta.config import Config
from inmanta.moduletool import ModuleTool
from moduletool.common import BadModProvider, install_project


def test_bad_checkout(modules_dir, modules_repo):
    coroot = os.path.join(modules_dir, "badproject")
    subprocess.check_output(
        ["git", "clone", os.path.join(modules_dir, "repos", "badproject")], cwd=modules_dir, stderr=subprocess.STDOUT
    )
    os.chdir(coroot)
    os.curdir = coroot
    Config.load_config()

    with pytest.raises(ModuleNotFoundException):
        ModuleTool().execute("install", [])


def test_bad_setup(modules_dir, modules_repo):
    coroot = os.path.join(modules_dir, "badprojectx")
    subprocess.check_output(
        ["git", "clone", os.path.join(modules_dir, "repos", "badproject"), coroot], cwd=modules_dir, stderr=subprocess.STDOUT
    )
    os.chdir(coroot)
    os.curdir = coroot
    Config.load_config()

    mod1 = os.path.join(coroot, "libs", "mod1")
    os.makedirs(mod1)
    subprocess.check_output(
        ["git", "clone", os.path.join(modules_dir, "repos", "mod2"), mod1], cwd=modules_dir, stderr=subprocess.STDOUT
    )

    with pytest.raises(ModuleNotFoundException):
        ModuleTool().execute("verify", [])


def test_complex_checkout(modules_dir, modules_repo):
    coroot = os.path.join(modules_dir, "testproject")
    subprocess.check_output(
        ["git", "clone", os.path.join(modules_dir, "repos", "testproject")], cwd=modules_dir, stderr=subprocess.STDOUT
    )
    os.chdir(coroot)
    os.curdir = coroot
    Config.load_config()

    ModuleTool().execute("install", [])
    expected = ["mod1", "mod2", "mod3", "mod6", "mod7"]
    for i in expected:
        dirname = os.path.join(coroot, "libs", i)
        assert os.path.exists(os.path.join(dirname, "signal"))
        assert not os.path.exists(os.path.join(dirname, "badsignal"))

    assert not os.path.exists(os.path.join(coroot, "libs", "mod5"))

    # test all tools, perhaps isolate to other test case
    ModuleTool().execute("list", [])
    ModuleTool().execute("update", [])
    ModuleTool().execute("status", [])
    ModuleTool().execute("push", [])


def test_for_git_failures(modules_dir, modules_repo):
    coroot = os.path.join(modules_dir, "testproject2")
    subprocess.check_output(
        ["git", "clone", os.path.join(modules_dir, "repos", "testproject"), "testproject2"],
        cwd=modules_dir,
        stderr=subprocess.STDOUT,
    )
    os.chdir(coroot)
    os.curdir = coroot
    Config.load_config()

    ModuleTool().execute("install", [])

    gp = module.gitprovider
    module.gitprovider = BadModProvider(gp, os.path.join(coroot, "libs", "mod6"))
    try:
        # test all tools, perhaps isolate to other test case
        ModuleTool().execute("install", [])
        ModuleTool().execute("list", [])
        ModuleTool().execute("update", [])
        ModuleTool().execute("status", [])
        ModuleTool().execute("push", [])
    finally:
        module.gitprovider = gp


def test_install_for_git_failures(modules_dir, modules_repo):
    coroot = os.path.join(modules_dir, "testproject3")
    subprocess.check_output(
        ["git", "clone", os.path.join(modules_dir, "repos", "testproject"), "testproject3"],
        cwd=modules_dir,
        stderr=subprocess.STDOUT,
    )
    os.chdir(coroot)
    os.curdir = coroot
    Config.load_config()

    gp = module.gitprovider
    module.gitprovider = BadModProvider(gp, os.path.join(coroot, "libs", "mod6"))
    try:
        with pytest.raises(ModuleNotFoundException):
            ModuleTool().execute("install", [])
    finally:
        module.gitprovider = gp


def test_for_repo_without_versions(modules_dir, modules_repo):
    coroot = os.path.join(modules_dir, "noverproject")
    subprocess.check_output(
        ["git", "clone", os.path.join(modules_dir, "repos", "noverproject")], cwd=modules_dir, stderr=subprocess.STDOUT
    )
    os.chdir(coroot)
    os.curdir = coroot
    Config.load_config()

    ModuleTool().execute("install", [])


def test_bad_dep_checkout(modules_dir, modules_repo):
    coroot = os.path.join(modules_dir, "baddep")
    subprocess.check_output(
        ["git", "clone", os.path.join(modules_dir, "repos", "baddep")], cwd=modules_dir, stderr=subprocess.STDOUT
    )
    os.chdir(coroot)
    os.curdir = coroot
    Config.load_config()

    with pytest.raises(CompilerException):
        ModuleTool().execute("install", [])


def test_master_checkout(modules_dir, modules_repo):
    coroot = install_project(modules_dir, "masterproject")

    ModuleTool().execute("install", [])

    dirname = os.path.join(coroot, "libs", "mod8")
    assert os.path.exists(os.path.join(dirname, "devsignal"))
    assert os.path.exists(os.path.join(dirname, "mastersignal"))


def test_dev_checkout(modules_dir, modules_repo):
    coroot = os.path.join(modules_dir, "devproject")
    subprocess.check_output(
        ["git", "clone", os.path.join(modules_dir, "repos", "devproject")], cwd=modules_dir, stderr=subprocess.STDOUT
    )
    os.chdir(coroot)
    os.curdir = coroot
    Config.load_config()

    ModuleTool().execute("install", [])

    dirname = os.path.join(coroot, "libs", "mod8")
    assert os.path.exists(os.path.join(dirname, "devsignal"))
    assert not os.path.exists(os.path.join(dirname, "mastersignal"))



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import os
import re
import shutil
import subprocess

import pytest
import yaml
from pkg_resources import parse_version

from inmanta import module
from inmanta.module import Project
from inmanta.moduletool import ModuleTool
from inmanta.parser import ParserException
from moduletool.common import add_file, commitmodule, install_project, make_module_simple, makeproject
from test_app_cli import app


def test_versioning():
    mt = ModuleTool()

    newversion = mt.determine_new_version(parse_version("1.2.3"), None, False, False, True, False)
    assert str(newversion) == "1.2.4"
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, False, True, False, False)
    assert str(newversion) == "1.3.0"
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, True, False, False, False)
    assert str(newversion) == "2.0.0"
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, True, True, False, False)
    assert newversion is None
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, True, False, True, False)
    assert newversion is None
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, True, True, True, False)
    assert newversion is None
    newversion = mt.determine_new_version(parse_version("1.2.3.dev025"), None, False, False, True, False)
    assert str(newversion) == "1.2.3"
    newversion = mt.determine_new_version(parse_version("1.2.3.dev025"), None, False, False, False, False)
    assert str(newversion) == "1.2.3"

    newversion = mt.determine_new_version(parse_version("1.2.3"), None, False, False, True, True)
    assert re.search("1.2.4.dev[0-9]+", str(newversion))
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, False, True, False, True)
    assert re.search("1.3.0.dev[0-9]+", str(newversion))
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, True, False, False, True)
    assert re.search("2.0.0.dev[0-9]+", str(newversion))
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, True, True, False, True)
    assert newversion is None
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, True, False, True, True)
    assert newversion is None
    newversion = mt.determine_new_version(parse_version("1.2.3"), None, True, True, True, True)
    assert newversion is None
    newversion = mt.determine_new_version(parse_version("1.2.3.dev025"), None, False, False, True, True)
    assert re.search("1.2.3.dev[0-9]+", str(newversion))
    newversion = mt.determine_new_version(parse_version("1.2.3.dev025"), None, False, False, False, True)
    assert re.search("1.2.3.dev[0-9]+", str(newversion))


def test_rewrite(tmpdir):
    module_path = tmpdir.join("mod").mkdir()
    model = module_path.join("model").mkdir()
    model.join("_init.cf").write("\n")

    module_yml = module_path.join("module.yml")
    module_yml.write(
        """
name: mod
license: ASL
version: 1.2
compiler_version: 2017.2
    """
    )

    mod = module.Module(None, module_path.strpath)

    assert mod.version == "1.2"
    assert mod.compiler_version == "2017.2"

    mod.rewrite_version("1.3.1")
    assert mod.version == "1.3.1"
    assert mod.compiler_version == "2017.2"


def test_module_corruption(modules_dir, modules_repo):
    mod9 = make_module_simple(modules_repo, "mod9", [("mod10", None)])
    add_file(mod9, "signal", "present", "third commit", version="3.3")
    add_file(mod9, "model/b.cf", "import mod9", "fourth commit", version="4.0")

    mod10 = make_module_simple(modules_repo, "mod10", [])
    add_file(mod10, "signal", "present", "a commit", version="3.3")
    # syntax error
    add_file(mod10, "model/_init.cf", "SomeInvalidThings", "a commit", version="3.5")
    # fix it
    add_file(mod10, "model/_init.cf", "", "a commit", version="3.6")
    add_file(mod10, "secondsignal", "import mod9", "b commit", version="4.0")
    add_file(mod10, "badsignal", "import mod9", "c commit", version="5.0")

    p9 = makeproject(modules_repo, "proj9", [("mod9", "==3.3"), ("mod10", "==3.3")], ["mod9"])
    commitmodule(p9, "first commit")

    # setup project
    proj = install_project(modules_dir, "proj9")
    app(["modules", "install"])
    print(os.listdir(proj))

    # unfreeze deps to allow update
    projectyml = os.path.join(proj, "project.yml")
    assert os.path.exists(projectyml)

    with open(projectyml, "r", encoding="utf-8") as fh:
        pyml = yaml.load(fh)

    pyml["requires"] = ["mod10 == 3.5"]

    with open(projectyml, "w", encoding="utf-8") as fh:
        yaml.dump(pyml, fh)

    # clear cache
    Project._project = None

    # attempt to update, mod10 is wrong, but only after the update
    app(["modules", "update"])

    with pytest.raises(ParserException):
        # clear cache
        Project._project = None
        # attempt to update, mod10 is wrong, can not be fixed
        app(["modules", "update"])

    # unfreeze deps to allow update
    pyml["requires"] = ["mod10 == 4.0"]

    with open(projectyml, "w", encoding="utf-8") as fh:
        yaml.dump(pyml, fh)

    # overwrite main to import unknown sub module
    main = os.path.join(proj, "main.cf")
    assert os.path.exists(main)

    with open(main, "w", encoding="utf-8") as fh:
        fh.write("import mod9::b")

    # clear cache
    Project._project = None

    # attempt to update
    app(["modules", "update"])

    # Additional output
    Project._project = None
    app(["modules", "list"])

    # Verify
    m9dir = os.path.join(proj, "libs", "mod9")
    assert os.path.exists(os.path.join(m9dir, "model", "b.cf"))
    m10dir = os.path.join(proj, "libs", "mod10")
    assert os.path.exists(os.path.join(m10dir, "secondsignal"))
    # should not be lastest version
    assert not os.path.exists(os.path.join(m10dir, "badsignal"))


@pytest.fixture(scope="function")
def module_without_tags(modules_repo):
    mod_no_tag = make_module_simple(modules_repo, "mod-no-tag")
    yield mod_no_tag
    shutil.rmtree(mod_no_tag)


@pytest.mark.parametrize(
    "dev, tag, version_tag_in_output", [(True, True, True), (True, False, False), (False, True, True), (False, False, True)],
)
def test_commit_no_tags(modules_dir, module_without_tags, dev, tag, version_tag_in_output):
    add_file(module_without_tags, "dummyfile", "Content", "Commit without tags", version="5.0", dev=dev, tag=tag)
    output = subprocess.check_output(["git", "tag", "-l"], cwd=module_without_tags, stderr=subprocess.STDOUT)
    assert ("5.0" in str(output)) is version_tag_in_output



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from functools import reduce
from itertools import chain
from typing import Callable, Dict, Iterator, List, Optional, Set, Tuple, Type

import pytest

import inmanta.ast.type as inmanta_type
import inmanta.compiler as compiler
from inmanta.ast import Namespace, RuntimeException
from inmanta.ast.blocks import BasicBlock
from inmanta.ast.entity import Entity
from inmanta.ast.statements import Statement
from inmanta.config import Config
from inmanta.execute.dataflow import (
    AssignableNode,
    AssignableNodeReference,
    AttributeNode,
    AttributeNodeReference,
    DataflowGraph,
    InstanceNode,
    NodeReference,
    ValueNode,
    VariableNodeReference,
)
from inmanta.execute.runtime import ExecutionContext, Resolver


def create_instance(
    graph: Optional[DataflowGraph] = None, entity: Optional[Entity] = None, statement: Optional[Statement] = None
) -> InstanceNode:
    responsible: Statement = statement if statement is not None else Statement()
    instance: InstanceNode = InstanceNode([])
    if graph is None:
        return instance
    return graph.own_instance_node_for_responsible(
        entity if entity is not None else Entity("DummyEntity", Namespace("dummy_namespace")), responsible, lambda: instance,
    )


@pytest.fixture(scope="function")
def graph() -> Iterator[DataflowGraph]:
    namespace: Namespace = Namespace("dummy_namespace")
    resolver: Resolver = Resolver(namespace, enable_dataflow_graph=True)
    block: BasicBlock = BasicBlock(namespace, [])
    xc: ExecutionContext = ExecutionContext(block, resolver)
    block.namespace.scope = xc

    yield DataflowGraph(resolver)


def get_dataflow_node(graph: DataflowGraph, name: str) -> AssignableNodeReference:
    """
        Returns a dataflow node for a graph by name. Name is allowed to have '.' for attribute nodes.
    """
    parts: List[str] = name.split(".")
    return reduce(lambda acc, part: AttributeNodeReference(acc, part), parts[1:], graph.resolver.get_dataflow_node(parts[0]))


class DataflowTestHelper:
    def __init__(self, snippetcompiler) -> None:
        self.snippetcompiler = snippetcompiler
        self._types: Dict[str, inmanta_type.Type] = {}
        self._namespace: Optional[Namespace] = None
        self._instances: Dict[str, InstanceNode] = {}
        self._tokens: List[str] = []

    def get_types(self) -> Dict[str, inmanta_type.Type]:
        return self._types

    def get_namespace(self) -> Namespace:
        assert self._namespace is not None, "Call compile before trying to access namespace"
        return self._namespace

    def get_graph(self) -> DataflowGraph:
        graph: Optional[DataflowGraph] = self.get_namespace().get_scope().dataflow_graph
        assert graph is not None
        return graph

    def compile(self, snippet: str, expected_error_type: Optional[Type[RuntimeException]] = None) -> None:
        def compile():
            self.snippetcompiler.setup_for_snippet(snippet)
            Config.set("compiler", "datatrace_enable", "true")
            (self._types, root_ns) = compiler.do_compile()
            self._namespace = root_ns.get_child("__config__")

        if expected_error_type is None:
            compile()
        else:
            try:
                compile()
                assert False, "Expected error: %s" % expected_error_type
            except expected_error_type as e:
                assert e.root_ns is not None
                root_ns = e.root_ns
                assert root_ns is not None
                self._namespace = root_ns.get_child("__config__")
            except Exception as e:
                if isinstance(e, AssertionError):
                    raise e
                assert False, "Expected %s, got %s" % (expected_error_type, e)

    def _consume_token_instance(self) -> Optional[str]:
        if self._tokens[0] != "<instance>":
            return None
        self._tokens.pop(0)
        instance_id: str = self._tokens.pop(0)
        if not instance_id.isalnum():
            raise Exception("Invalid syntax: expected instance identifier, got `%s`" % instance_id)
        return instance_id

    def _consume_token_attribute(self) -> Optional[str]:
        if len(self._tokens) == 0 or self._tokens[0] != ".":
            return None
        self._tokens.pop(0)
        attribute: str = self._tokens.pop(0)
        if not attribute.isalnum():
            raise Exception("Invalid syntax: expected attribute name, got `%s`" % attribute)
        return attribute

    def _consume_token_lhs(self) -> Callable[[List[NodeReference], Optional[str]], None]:
        node: AssignableNode
        instance_id: Optional[str] = self._consume_token_instance()
        if instance_id is not None:
            if instance_id not in self._instances:
                raise Exception("Parse error: bind instance_id `%s` to a n instance first by using it as a rhs." % instance_id)
            instance: InstanceNode = self._instances[instance_id]
            attribute_name: Optional[str] = self._consume_token_attribute()
            if attribute_name is None:
                raise Exception("Parse error: expected `. attribute_name`, got %s." % attribute_name)
            attribute: Optional[AttributeNode] = instance.get_attribute(attribute_name)
            assert attribute is not None
            node = attribute
        else:
            token: str = self._tokens.pop(0)
            if not token.isalnum():
                raise Exception("Invalid syntax: expected `variable_name` or `<instance> instance_id`, got `%s`" % token)
            node_ref: AssignableNodeReference = get_dataflow_node(self.get_graph(), token)
            assert isinstance(node_ref, VariableNodeReference)
            node = node_ref.node
        if self._consume_token_attribute() is not None:
            raise Exception("Syntax error: this simple language only supports attributes directly on instances in the lhs.")

        def continuation(rhs: List[NodeReference], instance_bind: Optional[str] = None) -> None:
            if instance_bind is not None:
                assert (
                    len(node.instance_assignments) == 1
                ), "This simple language only allows instance binding for a single instance in the rhs."
                instance_node: InstanceNode = node.instance_assignments[0].rhs.top_node()
                self._instances[instance_bind] = instance_node
                rhs.append(instance_node.reference())
            # test element equality: NodeReference does not define sort or hash so this is the only way
            actual_rhs: List[NodeReference] = [assignment.rhs for assignment in node.assignments()]
            for rhs_elem in rhs:
                assert rhs_elem in actual_rhs
                actual_rhs.remove(rhs_elem)
            assert len(actual_rhs) == 0

        return continuation

    def _consume_token_edge(self) -> None:
        token: str = self._tokens.pop(0)
        if token != "->":
            raise Exception("Invalid syntax: expected `->`, got `%s`" % token)

    def _consume_token_rhs_element(self) -> Tuple[Optional[NodeReference], Optional[str]]:
        instance_id: Optional[str] = self._consume_token_instance()
        if instance_id is not None:
            if self._consume_token_attribute() is not None:
                raise Exception("Syntax error: this simple language only supports attributes directly on instances in the lhs.")
            if instance_id not in self._instances:
                return (None, instance_id)
            return (self._instances[instance_id].reference(), None)
        else:
            token: str = self._tokens.pop(0)
            if not token.isalnum():
                raise Exception(
                    "Invalid syntax: expected `variable_name [. attr [...]]` or `<instance> instance_id`, got `%s`" % token
                )
            try:
                return (ValueNode(int(token)).reference(), None)
            except ValueError:
                node_ref: AssignableNodeReference = get_dataflow_node(self.get_graph(), token)
                assert isinstance(node_ref, VariableNodeReference)
                attribute_name: Optional[str] = self._consume_token_attribute()
                while attribute_name is not None:
                    node_ref = AttributeNodeReference(node_ref, attribute_name)
                    attribute_name = self._consume_token_attribute()
                return (node_ref, None)

    def _consume_token_rhs(self, continuation: Callable[[List[NodeReference], Optional[str]], None]) -> None:
        nodes: List[NodeReference] = []
        instance_bind: Optional[str] = None

        def consume_node(instance_bind: Optional[str] = None) -> Optional[str]:
            (node, instance_id) = self._consume_token_rhs_element()
            if instance_id is not None:
                if instance_bind is not None:
                    raise Exception("Parse error: this simple language only allows a single instance binding in the rhs.")
                instance_bind = instance_id
            if node is not None:
                nodes.append(node)
            return instance_bind

        if self._tokens[0] == "[":
            self._tokens.pop(0)
            while self._tokens[0] != "]":
                instance_bind = consume_node(instance_bind)
            self._tokens.pop(0)
        else:
            instance_bind = consume_node(instance_bind)
        continuation(nodes, instance_bind)

    def verify_graphstring(self, graphstring: str) -> None:
        """
            Verifies that the graphstring corresponds with the graph. Syntax for the graphstring:
                graph_string: empty
                    | graphstring_rule graph_string
                    ;
                graphstring_rule : lhs `->` rhs ;
                lhs : var_name
                    | instance_ref
                    | instance_ref `.` attr_name
                    ;
                rhs : rhs_element
                    | `[` rhs_list `]`
                    ;
                rhs_list : empty
                    | rhs_element rhs_list
                    ;
                rhs_element : var_name
                    | int
                    | var_attr
                    | `<instance>` instance_id
                    ;
                var_attr : var_name `.` attr_name`
                    | var_name `.` var_attr
                    ;
            Where instance_id is a string that is bound when it occurs in the right. Once bound the same id can be used to refer
            to it in later rules.
            Spaces are required between all subsequent tokens.
        """
        self._tokens = graphstring.split()
        while len(self._tokens) > 0:
            continuation: Callable[[List[NodeReference], Optional[str]], None] = self._consume_token_lhs()
            self._consume_token_edge()
            self._consume_token_rhs(continuation)

    def verify_leaves(self, leaves: Dict[str, Set[str]]) -> None:
        """
            Verifies that the leaves correspond with the graph's leaves.
            :param leaves: dict with variable names as keys and a set of leaves for each variable as values.
                The variable and leaves are allowed to be attributes.
        """
        for key, value in leaves.items():
            lhs: AssignableNodeReference = get_dataflow_node(self.get_graph(), key)
            rhs: Set[AssignableNode] = set(chain.from_iterable(get_dataflow_node(self.get_graph(), v).nodes() for v in value))
            assert set(lhs.leaf_nodes()) == rhs


@pytest.fixture(scope="function")
def dataflow_test_helper(snippetcompiler):
    return DataflowTestHelper(snippetcompiler)



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import inmanta.compiler as compiler
from inmanta.util import get_compiler_version


def test_compiler_installed():
    """
    If this test fails, make sure inmanta is installed. For example with pip install -e .
    """
    assert get_compiler_version() is not None


def test_str_on_instance_pos(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import std

entity Hg:
end

Hg.hosts [0:] -- std::Host

implement Hg using std::none

hg = Hg()

for i in [1,2,3]:
 hg.hosts = std::Host(name="Test{{i}}", os=std::unix)
end


for i in hg.hosts:
    std::ConfigFile(host=i, path="/fx", content="")
end
"""
    )
    (types, _) = compiler.do_compile()
    files = types["std::File"].get_all_instances()
    assert len(files) == 3


def test_str_on_instance_neg(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import std

entity Hg:
end

Hg.hosts [0:] -- std::Host

implement Hg using std::none

hg = Hg()

for i in [1,2,3]:
 hg.hosts = std::Host(name="Test", os=std::unix)
end


for i in hg.hosts:
    std::ConfigFile(host=i, path="/fx", content="")
end
"""
    )
    (types, _) = compiler.do_compile()
    files = types["std::File"].get_all_instances()
    assert len(files) == 1


def test_implements_inheritance(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    string a
end

entity TestC extends Test:
end

implementation test for Test:
    self.a = "xx"
end


implement TestC using parents
implement TestC using std::none, parents
implement TestC using std::none
implement Test using test

a = TestC()
"""
    )
    (_, scopes) = compiler.do_compile()

    root = scopes.get_child("__config__")
    assert "xx" == root.lookup("a").get_value().lookup("a").get_value()


def test_keyword_excn(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
       index = ""
""",
        "Syntax error: invalid identifier, index is a reserved keyword ({dir}/main.cf:2:8)",
    )


def test_keyword_excn2(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
       implementation index for std::Entity:
       end
""",
        "Syntax error: invalid identifier, index is a reserved keyword ({dir}/main.cf:2:23)",
    )


def test_keyword_excn3(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
       implementation aaa for index::Entity:
       end
""",
        "Syntax error: invalid identifier, index is a reserved keyword ({dir}/main.cf:2:31)",
    )


def test_cid_excn(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
       entity test:
       end
""",
        "Syntax error: Invalid identifier: Entity names must start with a capital ({dir}/main.cf:2:15)",
    )


def test_cid_excn2(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
       entity Test extends test:
       end
""",
        "Syntax error: Invalid identifier: Entity names must start with a capital ({dir}/main.cf:2:28)",
    )


def test_bad_var(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
        a=b
""",
        "variable b not found (reported in a = b ({dir}/main.cf:2))",
    )


def test_bad_type(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
entity Test1:
    string a
end

Test1(a=3)
""",
        """Could not set attribute `a` on instance `__config__::Test1 (instantiated at {dir}/main.cf:6)` """
        """(reported in Construct(Test1) ({dir}/main.cf:6))
caused by:
  Invalid value '3', expected String (reported in Construct(Test1) ({dir}/main.cf:6))""",
    )


def test_bad_type_2(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
import std

entity Test1:
    string a
end

implement Test1 using std::none

t1 = Test1()
t1.a=3
""",
        """Could not set attribute `a` on instance `__config__::Test1 (instantiated at {dir}/main.cf:10)` (reported in t1.a = 3 ({dir}/main.cf:11))
caused by:
  Invalid value '3', expected String (reported in t1.a = 3 ({dir}/main.cf:11))""",  # noqa: E501
    )


def test_value_set_twice(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
test = "a"
test = "b"
""",
        """value set twice:
\told value: a
\t\tset at {dir}/main.cf:2
\tnew value: b
\t\tset at {dir}/main.cf:3
 (reported in test = 'b' ({dir}/main.cf:3))""",
    )



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from compiler.dataflow.conftest import DataflowTestHelper

import pytest


@pytest.mark.parametrize("bidirectional", [True, False])
@pytest.mark.parametrize("inherit_relation", [True, False])
@pytest.mark.parametrize("assign_first", [True, False])
def test_dataflow_model_relation(
    dataflow_test_helper: DataflowTestHelper, bidirectional: bool, inherit_relation: bool, assign_first: bool
) -> None:
    relation_stmt: str = "%s.b [1] -- B%s" % ("AParent" if inherit_relation else "A", ".a [1]" if bidirectional else "")
    dataflow_test_helper.compile(
        """
entity AParent:
end
implement AParent using std::none

entity A extends AParent:
end
implement A using std::none

entity B:
end
implement B using std::none

%s

a = A()
b = B()

a.b = b

%s
        """
        % ("" if assign_first else relation_stmt, relation_stmt if assign_first else "")
    )
    bidirectional_rule: str = "<instance> b . a -> <instance> a"
    dataflow_test_helper.verify_graphstring(
        """
a -> <instance> a
b -> <instance> b

<instance> a . b -> b
%s
        """
        % (bidirectional_rule if bidirectional else ""),
    )
    if not bidirectional:
        with pytest.raises(AssertionError):
            dataflow_test_helper.verify_graphstring(bidirectional_rule)
    dataflow_test_helper.verify_leaves({"a": {"a"}, "b": {"b"}, "a.b": {"b"}, "b.a": {"b.a"}})


def test_dataflow_model_assignment_to_relation(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
entity X:
end

entity U:
end

entity V:
    number n
end

X.u [1] -- U
U.v [1] -- V

implement X using std::none
implement U using std::none
implement V using std::none

n = 42

x = X()
x.u = U()
x.u.v = V()
x.u.v.n = n
        """,
    )
    dataflow_test_helper.verify_graphstring(
        """
x -> <instance> x
<instance> x . u -> <instance> u
<instance> u . v -> <instance> v
<instance> v . n -> n
n -> 42
        """,
    )
    dataflow_test_helper.verify_leaves({"x.u.v.n": {"n"}})


def test_dataflow_model_assignment_from_relation(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
entity U:
end

entity V:
    number n
end

U.v [1] -- V

implement U using std::none
implement V using std::none

n = 42

u = U(v = v)
v = V(n = n)

uvn = u.v.n
        """,
    )
    dataflow_test_helper.verify_graphstring(
        """
n -> 42
u -> <instance> u
v -> <instance> v

<instance> u . v -> v
<instance> v . n -> n

uvn -> u . v . n
        """,
    )
    dataflow_test_helper.verify_leaves({"n": {"n"}, "u": {"u"}, "v": {"v"}, "u.v": {"v"}, "v.n": {"n"}, "uvn": {"n"}})


def test_dataflow_model_index(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
entity A:
    number n
    number k
    number l
end

index A(n)

implement A using std::none

x = A(n = 42, k = 0)
y = A(n = 42, l = 1)
        """,
    )
    dataflow_test_helper.verify_graphstring(
        """
x -> <instance> x
y -> <instance> y

<instance> x . n -> [ 42 42 ]
<instance> y . n -> [ 42 42 ]

<instance> x . k -> 0
<instance> y . k -> 0

<instance> x . l -> 1
<instance> y . l -> 1
        """,
    )


def test_dataflow_model_default_attribute(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
entity A:
    number n = 42
end

implement A using std::none

x = A()
y = A(n = 0)
        """,
    )
    dataflow_test_helper.verify_graphstring(
        """
x -> <instance> x
y -> <instance> y

<instance> x . n -> 42
<instance> y . n -> 0
        """,
    )


@pytest.mark.parametrize("refer_out", [True, False])
def test_dataflow_model_implementation(dataflow_test_helper: DataflowTestHelper, refer_out: bool) -> None:
    dataflow_test_helper.compile(
        """
entity A:
    number n
end

implementation i for A:
    self.n = %s
end

implement A using i

nn = 42
x = A()
        """
        % ("nn" if refer_out else 42),
    )
    dataflow_test_helper.verify_graphstring(
        """
nn -> 42
x -> <instance> x
<instance> x . n -> %s
        """
        % ("nn" if refer_out else 42),
    )


def test_dataflow_model_unsupported_bidirectional_doesnt_crash(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
entity A:
end

entity B:
end

implement A using std::none
implement B using std::none

A.b [0:] -- B.a [0:]

a = A()
# Lists are not supported yet. Mustn't crash on trying to model the other side of the bidirectional relation
a.b = [B(), B()]
        """,
    )



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from itertools import chain
from typing import Iterable, List, Optional

from inmanta.ast import Locatable, NotFoundException
from inmanta.ast.statements import Statement
from inmanta.execute.dataflow import (
    AssignableNode,
    AssignableNodeReference,
    Assignment,
    AttributeNodeReference,
    DataflowGraph,
    Equivalence,
    InstanceAttributeNodeReference,
    InstanceNode,
    InstanceNodeReference,
    VariableNodeReference,
)
from inmanta.execute.runtime import Instance


class DataTraceRenderer:
    """
        Renderer for the data trace of an assignable node. The data trace shows all data paths to the node as well as dynamic
        context (such as implementations) where applicable. The main entrypoint is the render() method.
    """

    @classmethod
    def render(cls, node_ref: AssignableNodeReference) -> str:
        return "\n".join(cls._render_reference(node_ref, tree_root=True)) + "\n"

    @classmethod
    def _render_reference(cls, node_ref: AssignableNodeReference, tree_root: bool = False) -> List[str]:
        """
            Renders the data trace for all nodes a reference refers to.
        """
        result: List[str] = []
        if tree_root:
            result.append(repr(node_ref))
        if isinstance(node_ref, AttributeNodeReference):
            result.append("SUBTREE for %s:" % node_ref.instance_var_ref)
            result += cls._indent(cls._render_reference(node_ref.instance_var_ref))
        if isinstance(node_ref, InstanceAttributeNodeReference):
            result.append("SUBTREE for %s:" % node_ref.instance)
            result += cls._indent(cls._render_instance(node_ref.instance))
        for node in node_ref.nodes():
            result += cls._render_node(node)
        return result

    @classmethod
    def _render_node(cls, node: AssignableNode) -> List[str]:
        """
            Renders the data trace for an assignable node. Shows information about:
                - the node's parent instance, if it is an attribute node
                - the node's equivalence
                - assignments to the node:
                    - right hand side
                    - responsible
                    - the dynamic context it lives in, if any
            Recurses on the assignment's right hand side.

            :param tree_root: indicates whether this node is the root of the data trace tree. Behaviour for non-root nodes is
                slightly different in order to prevent output duplication.
        """
        result: List[str] = []
        result += cls._render_equivalence(node.equivalence)
        assignments: List[Assignment] = list(
            chain(
                node.equivalence.external_assignable_assignments(),
                node.equivalence.instance_assignments(),
                node.equivalence.value_assignments(),
            )
        )
        nb_assignments: int = len(assignments)
        for i, assignment in enumerate(assignments):
            last: bool = i == nb_assignments - 1
            subblock: List[str] = []

            subblock += cls._render_assignment(assignment)
            subblock += cls._render_implementation_context(assignment.context)

            if isinstance(assignment.rhs, InstanceNodeReference):
                subblock += cls._render_instance(assignment.rhs.top_node())
            if isinstance(assignment.rhs, AssignableNodeReference):
                subblock += cls._render_reference(assignment.rhs)

            result += cls._branch(subblock, last)
        return result

    @classmethod
    def _prefix_line(cls, prefix: str, line: str) -> str:
        """
            Prefixes a line.
        """
        if line == "":
            return prefix.rstrip()
        return prefix + line

    @classmethod
    def _prefix(cls, prefix: str, lines: Iterable[str]) -> List[str]:
        """
            Prefixes lines.
        """
        return [cls._prefix_line(prefix, line) for line in lines]

    @classmethod
    def _branch(cls, lines: List[str], last: Optional[bool] = False) -> List[str]:
        """
            Renders a branch in the tree.

            :param last: True iff this is the last branch on this level.
        """
        if len(lines) == 0:
            return []
        branch_prefix: str = ("└" if last else "├") + "── "
        block_prefix: str = (" " if last else "│") + " " * 3
        result: List[str] = cls._prefix(branch_prefix, lines[0:1])
        result += cls._prefix(block_prefix, lines[1:])
        return result

    @classmethod
    def _indent(cls, lines: Iterable[str]) -> List[str]:
        """
            Indents lines.
        """
        return cls._prefix(" " * 4, lines)

    @classmethod
    def _render_implementation_context(cls, context: DataflowGraph) -> List[str]:
        """
            Renders information about the dynamic implementation context, if it exists.
        """
        # roundabout way to detect encapsulating context, may lead to false positives, see #1937
        try:
            context.resolver.lookup("self")
        except NotFoundException:
            return []
        result: List[str] = []
        var_node: AssignableNodeReference = context.resolver.get_dataflow_node("self")
        if (
            isinstance(var_node, VariableNodeReference)
            and len(var_node.node.instance_assignments) == 1
            and isinstance(var_node.node.instance_assignments[0].responsible, Instance)
        ):
            instance_node: InstanceNode = var_node.node.instance_assignments[0].rhs.top_node()
            result.append("IN IMPLEMENTATION WITH self = %s" % instance_node)
            result += cls._indent(cls._render_constructor(instance_node))
        return result

    @classmethod
    def _render_constructor(cls, instance: InstanceNode) -> List[str]:
        """
            Renders information about the construction of an instance node:
                - constructor statement
                - lexical position
                - dynamic context it lives in, if any
        """
        result: List[str] = []
        if instance.responsible is not None:
            result += [
                "CONSTRUCTED BY `%s`" % instance.responsible.pretty_print(),
                "AT %s" % instance.responsible.get_location(),
            ]
        if instance.context is not None:
            result += cls._render_implementation_context(instance.context)
        return result

    @classmethod
    def _render_instance(cls, instance: InstanceNode) -> List[str]:
        """
            Renders information about an instance node:
                - construction information
                - index matches and their construction information
        """
        result: List[str] = []
        result += cls._render_constructor(instance)
        for index_node in instance.get_all_index_nodes():
            if index_node is instance:
                continue
            result += [
                "",
                "INDEX MATCH: `%s`" % index_node,
            ]

            subblock: List[str] = []
            subblock += cls._render_constructor(index_node)

            result += cls._indent(subblock)
        return result

    @classmethod
    def _render_assignment(cls, assignment: Assignment) -> List[str]:
        """
            Renders information about an assignment in the dataflow graph.
        """
        responsible: "Locatable" = assignment.responsible
        return [
            "%s" % assignment.rhs,
            "SET BY `%s`" % (responsible.pretty_print() if isinstance(responsible, Statement) else responsible),
            "AT %s" % responsible.get_location(),
        ]

    @classmethod
    def _render_equivalence(cls, equivalence: Equivalence) -> List[str]:
        """
            Renders information about an equivalence unless trivial. Shows the equivalence's members and the responsible
            assignments.
        """
        if len(equivalence.nodes) > 1:
            # sort output for consistency
            return [
                "EQUIVALENT TO {%s} DUE TO STATEMENTS:" % ", ".join(sorted(repr(n) for n in equivalence.nodes)),
                *cls._indent(
                    [
                        "`%s` AT %s" % resp_loc
                        for resp_loc in sorted(
                            (
                                (
                                    assignment.responsible.pretty_print()
                                    if isinstance(assignment.responsible, Statement)
                                    else str(assignment.responsible),
                                    str(assignment.responsible.get_location()),
                                )
                                for assignment in equivalence.interal_assignments()
                            ),
                            key=lambda t: tuple(reversed(t)),
                        )
                    ]
                ),
            ]
        return []



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from compiler.dataflow.conftest import DataflowTestHelper, get_dataflow_node
from typing import Dict, List, Optional

import pytest

import inmanta.ast.type as inmanta_type
from inmanta.ast import DoubleSetException, NotFoundException, RuntimeException
from inmanta.ast.entity import Entity
from inmanta.ast.statements import Literal
from inmanta.ast.statements.assign import Assign, SetAttribute
from inmanta.ast.variables import Reference
from inmanta.execute.dataflow import (
    AssignableNode,
    AssignableNodeReference,
    Assignment,
    AttributeNode,
    DataflowGraph,
    ValueNode,
    ValueNodeReference,
    VariableNodeReference,
)
from inmanta.execute.runtime import Instance, ResultVariable, Typeorvalue


def test_dataflow_model_primitive_assignment_responsible(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
x = 42
        """,
    )
    graph: DataflowGraph = dataflow_test_helper.get_graph()
    x: AssignableNodeReference = get_dataflow_node(graph, "x")
    assert isinstance(x, VariableNodeReference)
    assert len(x.node.value_assignments) == 1
    assignment: Assignment[ValueNodeReference] = x.node.value_assignments[0]
    assert isinstance(assignment.responsible, Assign)
    assert assignment.responsible.name == "x"
    assert isinstance(assignment.responsible.value, Literal)
    assert assignment.responsible.value.value == 42
    assert assignment.context == graph


def test_dataflow_model_primitive_double_assignment_responsible(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
x = 42
x = 0
        """,
        DoubleSetException,
    )
    graph: DataflowGraph = dataflow_test_helper.get_graph()
    x: AssignableNodeReference = get_dataflow_node(graph, "x")
    assert isinstance(x, VariableNodeReference)
    assignments: List[Assignment] = x.node.value_assignments
    assert len(assignments) == 2
    zero_index: int = [assignment.rhs for assignment in assignments].index(ValueNode(0).reference())
    for i, assignment in enumerate(assignments):
        value: int = 0 if i == zero_index else 42
        assert assignment.context == graph
        assert isinstance(assignment.responsible, Assign)
        assert assignment.responsible.name == "x"
        assert isinstance(assignment.responsible.value, Literal)
        assert assignment.responsible.value.value == value


def test_dataflow_model_variable_assignment_responsible(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
x = y
y = 42
        """,
    )
    graph: DataflowGraph = dataflow_test_helper.get_graph()
    x: AssignableNodeReference = get_dataflow_node(graph, "x")
    assert isinstance(x, VariableNodeReference)
    assert len(x.node.assignable_assignments) == 1
    assignment: Assignment[AssignableNodeReference] = x.node.assignable_assignments[0]
    assert isinstance(assignment.responsible, Assign)
    assert assignment.responsible.name == "x"
    assert isinstance(assignment.responsible.value, Reference)
    assert assignment.responsible.value.name == "y"
    assert assignment.context == graph


def test_dataflow_model_attribute_assignment_responsible(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
entity Test:
    number n
end
implement Test using std::none

x = Test()
x.n = 42
        """
    )
    graph: DataflowGraph = dataflow_test_helper.get_graph()
    x: AssignableNodeReference = get_dataflow_node(graph, "x")
    assert isinstance(x, VariableNodeReference)
    assert len(x.node.instance_assignments) == 1
    n: Optional[AttributeNode] = x.node.instance_assignments[0].rhs.node().get_attribute("n")
    assert n is not None
    assert len(n.value_assignments) == 1
    assignment: Assignment[ValueNodeReference] = n.value_assignments[0]
    assert isinstance(assignment.responsible, SetAttribute)
    assert assignment.responsible.instance.name == "x"
    assert assignment.responsible.attribute_name == "n"
    assert isinstance(assignment.responsible.value, Literal)
    assert assignment.responsible.value.value == 42
    assert assignment.context == graph


def test_dataflow_model_simple_assignment(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
x = 42
        """,
    )
    dataflow_test_helper.verify_graphstring(
        """
x -> 42
        """,
    )
    dataflow_test_helper.verify_leaves({"x": {"x"}})


def test_dataflow_model_variable_assignment(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
x = y
x = y
y = 42
        """,
    )
    dataflow_test_helper.verify_graphstring(
        """
x -> [ y y ]
y -> 42
        """,
    )
    dataflow_test_helper.verify_leaves({"x": {"y"}, "y": {"y"}})


@pytest.mark.parametrize("same_value", [True, False])
def test_dataflow_model_variable_assignment_double(dataflow_test_helper: DataflowTestHelper, same_value: bool) -> None:
    x_value: int = 42 if same_value else 0
    dataflow_test_helper.compile(
        """
x = y
x = y
y = 42
x = %d
        """
        % x_value,
        None if same_value else DoubleSetException,
    )
    dataflow_test_helper.verify_graphstring(
        """
x -> [ y y %s ]
y -> 42
        """
        % x_value,
    )
    dataflow_test_helper.verify_leaves({"x": {"x", "y"}, "y": {"y"}})


def test_dataflow_model_unassigned_dependency_error(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
x = y
y = z
        """,
        NotFoundException,
    )
    dataflow_test_helper.verify_graphstring(
        """
x -> y
y -> z
        """,
    )
    dataflow_test_helper.verify_leaves({"x": {"z"}, "y": {"z"}})


@pytest.mark.parametrize("attr_init", [True, False])
def test_dataflow_model_assignment_from_attribute(dataflow_test_helper: DataflowTestHelper, attr_init: bool) -> None:
    dataflow_test_helper.compile(
        """
entity A:
    number n
end
implement A using std::none

n = 42
x = A(%s)

nn = x.n
        """
        % ("n = n" if attr_init else ""),
        None if attr_init else RuntimeException,
    )
    dataflow_test_helper.verify_graphstring(
        """
x -> <instance> 0
n -> 42
nn -> x . n
%s
        """
        % ("<instance> 0 . n -> n" if attr_init else ""),
    )
    dataflow_test_helper.verify_leaves({"nn": {"n"}} if attr_init else {"nn": {"x.n"}})


def test_dataflow_model_assignment_outside_constructor(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
entity A:
    number n
end
implement A using std::none

n = 42

x = A()
x.n = n
        """,
    )
    dataflow_test_helper.verify_graphstring(
        """
n -> 42
x -> <instance> 0
<instance> 0 . n -> n
        """,
    )
    dataflow_test_helper.verify_leaves({"x.n": {"n"}})


def test_dataflow_model_no_leaf_error(dataflow_test_helper: DataflowTestHelper) -> None:
    """
        Verify that tentative instance propagation does not crash when rhs is a reference with no leaf nodes.
    """
    dataflow_test_helper.compile(
        """
n = x.n
        """,
        NotFoundException,
    )


def test_dataflow_model_result_variable(dataflow_test_helper: DataflowTestHelper) -> None:
    dataflow_test_helper.compile(
        """
entity A:
    number n
end

index A(n)

A.other [0:1] -- B

entity B:
end

implement A using std::none
implement B using std::none

x = A(n = 4)
x.other = B()

y = A(n = 4)
        """,
    )

    def assert_rv(result_variable: ResultVariable) -> None:
        node_ref: Optional[AssignableNodeReference] = result_variable.get_dataflow_node()
        assert node_ref is not None
        nodes: List[AssignableNode] = list(node_ref.nodes())
        assert len(nodes) == 1
        assert nodes[0].result_variable is result_variable

    x: Typeorvalue = dataflow_test_helper.get_namespace().lookup("x")
    assert isinstance(x, ResultVariable)
    assert_rv(x)

    types: Dict[str, inmanta_type.Type] = dataflow_test_helper.get_types()
    a_entity_str: str = "__config__::A"
    assert a_entity_str in types
    a_entity: inmanta_type.Type = types[a_entity_str]
    assert isinstance(a_entity, Entity)
    a_instances: List[Instance] = a_entity.get_all_instances()
    assert len(a_instances) == 1
    for attr in ["n", "other"]:
        assert_rv(a_instances[0].get_attribute(attr))



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import sys
import threading

from tornado import gen
from tornado.ioloop import IOLoop

import inmanta.const
from inmanta.app import setup_signal_handlers

inmanta.const.SHUTDOWN_GRACE_IOLOOP = 1
inmanta.const.SHUTDOWN_GRACE_HARD = 2


class MiniApp:
    def __init__(self):
        self.running = True
        self.lock = threading.Semaphore(0)

    @gen.coroutine
    def stop(self):
        print("STOP")
        self.running = False
        yield gen.sleep(0.2)

    @gen.coroutine
    def run(self):
        i = 0
        while self.running:
            yield asyncio.sleep(0.1)
            print(i)
            i += 1
        print("DONE")

    @gen.coroutine
    def bad_run(self):
        i = 0
        while self.running:
            print(i)
            i += 1
            yield asyncio.sleep(0.1)
            with self.lock:
                print(i)
        print("DONE")


if __name__ == "__main__":
    print("Start")
    a = MiniApp()
    setup_signal_handlers(a.stop)
    if "bad" in sys.argv:
        IOLoop.current().add_callback(a.bad_run)
    else:
        IOLoop.current().add_callback(a.run)
    IOLoop.current().start()
    print("SHUTDOWN COMPLETE")
    sys.exit(0)



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import re
import shutil
from compiler.dataflow.conftest import DataflowTestHelper
from functools import total_ordering
from itertools import chain
from typing import Callable, List, Optional, Tuple

import pytest

from inmanta.ast import Locatable, Namespace
from inmanta.ast.entity import Entity
from inmanta.execute.dataflow import AssignableNode, AttributeNode, InstanceNode, Node, ValueNode
from inmanta.execute.runtime import Instance

try:
    from inmanta.execute.dataflow.graphic import GraphicGraph
except Exception:
    pytest.skip(
        "skipping graphic tests because graphviz package not installed. Run `pip install graphviz` to resolve this.",
        allow_module_level=True,
    )

if shutil.which("fdp") is None:
    pytest.skip(
        "skipping graphic tests because graphviz fdp executable was not found in your $PATH."
        " Install your distribution's graphviz package to resolve this.",
        allow_module_level=True,
    )


def dot_stringify(obj: object) -> str:
    return re.sub(r"\W+", "", str(obj))


class LocationBasedGraphicGraph(GraphicGraph):
    def node_key(self, node: Node) -> str:
        if isinstance(node, InstanceNode):
            assert node.responsible is not None
            return "cluster_instance_%s_%s" % (dot_stringify(node), dot_stringify(node.responsible.location))
        elif isinstance(node, ValueNode):
            return "value_%s" % dot_stringify(node.value)
        elif isinstance(node, AttributeNode):
            assert node.instance.responsible is not None
            return "attribute_%s_on_%s_%s" % (
                dot_stringify(node.name),
                dot_stringify(node.instance),
                dot_stringify(node.instance.responsible.location),
            )
        elif isinstance(node, AssignableNode):
            assert node.result_variable is not None
            locatable: Locatable
            if node.result_variable.location is not None:
                locatable = node.result_variable
            else:
                # __self__ node
                value: object = node.result_variable.get_value()
                assert isinstance(value, Instance)
                locatable = value
            assert locatable.location is not None
            return "assignable_%s_%s" % (dot_stringify(node.name), dot_stringify(locatable.location))
        assert False


GraphicAsserter = Callable[[str, str], None]


@total_ordering
class DotSource:
    """
        Represents DOT source semi-structurally. Sorts the lines before comparing to deal with nondeterministism
        during construction.
    """

    def __init__(self, lines: List[str], subgraphs: List[Tuple[str, "DotSource"]]) -> None:
        self.lines: List[str] = sorted(lines)
        self.subgraphs: List[Tuple[str, DotSource]] = sorted(subgraphs)

    @classmethod
    def parse(cls, source: str) -> "DotSource":
        (lines, subgraphs, rest) = cls._get_lines_and_subgraphs(source.split("\n"))
        assert len(rest) == 0
        return DotSource(lines, subgraphs)

    @classmethod
    def _get_lines_and_subgraphs(cls, lines: List[str]) -> Tuple[List[str], List[Tuple[str, "DotSource"]], List[str]]:
        if len(lines) == 0:
            return ([], [], [])
        (head, tail) = (lines[0], lines[1:])
        if head.endswith("{"):
            (sublines, subsubs, rest) = cls._get_lines_and_subgraphs(tail)
            (rlines, rsubs, rrest) = cls._get_lines_and_subgraphs(rest)
            return (rlines, [(head, DotSource(sublines, subsubs)), *rsubs], rrest)
        if head.endswith("}"):
            return ([head], [], tail)
        (tail_lines, tail_subs, tail_rest) = cls._get_lines_and_subgraphs(tail)
        return ([head, *tail_lines], tail_subs, tail_rest)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DotSource):
            return NotImplemented
        return self.lines == other.lines and self.subgraphs == other.subgraphs

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, DotSource):
            return NotImplemented
        return (self.lines, self.subgraphs) < (other.lines, other.subgraphs)


@pytest.fixture(scope="function")
def graphic_asserter(dataflow_test_helper: DataflowTestHelper) -> Callable[[str, str], None]:
    def asserter(model: str, expected: str, view: Optional[bool] = False) -> None:
        dataflow_test_helper.compile(model)
        namespace: Namespace = dataflow_test_helper.get_namespace()
        entities: List[Entity] = [
            tp for tp in dataflow_test_helper.get_types().values() if isinstance(tp, Entity) if tp.namespace is namespace
        ]
        graphic: GraphicGraph = LocationBasedGraphicGraph()
        for instance in chain.from_iterable(entity.get_all_instances() for entity in entities):
            assert instance.instance_node is not None
            graphic.add_node(instance.instance_node.top_node())
        for result_variable in namespace.get_scope().slots.values():
            for node in result_variable.get_dataflow_node().nodes():
                graphic.add_node(node)
        if view:
            graphic.view()
        print(graphic.digraph.source.strip())
        assert DotSource.parse(graphic.digraph.source.strip()) == DotSource.parse(
            expected.strip().format(dir=dot_stringify(dataflow_test_helper.snippetcompiler.project_dir))
        )

    return asserter


def test_dataflow_graphic_simple_assignment(graphic_asserter: GraphicAsserter) -> None:
    graphic_asserter(
        """
x = 0
        """,
        """
digraph {{
\tassignable_x_{dir}maincf2 [label=x shape=ellipse]
\tvalue_0 [label=0 shape=diamond]
\tassignable_x_{dir}maincf2 -> value_0
}}
        """,
    )


def test_dataflow_graphic_assignment_loop(graphic_asserter: GraphicAsserter) -> None:
    graphic_asserter(
        """
x = y
y = z
z = x

y = 42
        """,
        """
digraph {{
\tassignable_x_{dir}maincf25 [label=x shape=ellipse]
\tassignable_y_{dir}maincf6 [label=y shape=ellipse]
\tassignable_z_{dir}maincf45 [label=z shape=ellipse]
\tassignable_z_{dir}maincf45 -> assignable_x_{dir}maincf25
\tassignable_y_{dir}maincf6 -> assignable_z_{dir}maincf45
\tvalue_42 [label=42 shape=diamond]
\tassignable_y_{dir}maincf6 -> value_42
\tassignable_x_{dir}maincf25 -> assignable_y_{dir}maincf6
}}

        """,
    )


def test_dataflow_graphic_instance(graphic_asserter: GraphicAsserter) -> None:
    graphic_asserter(
        """
entity A:
    number l
    number m
    number n
end

implement A using std::none


x = A(n = 42)
x.m = 0

y = x
y.l = 1


u = A(l = y.l, m = x.m, n = 2)
        """,
        """
digraph {{
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf11 {{
\t\tlabel=A
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf11 {{
\t\tattribute_l_on___config__Ainstance_{dir}maincf11 [label=l shape=ellipse]
\t}}
\tvalue_1 [label=1 shape=diamond]
\tattribute_l_on___config__Ainstance_{dir}maincf11 -> value_1
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf11 {{
\t\tattribute_m_on___config__Ainstance_{dir}maincf11 [label=m shape=ellipse]
\t}}
\tvalue_0 [label=0 shape=diamond]
\tattribute_m_on___config__Ainstance_{dir}maincf11 -> value_0
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf11 {{
\t\tattribute_n_on___config__Ainstance_{dir}maincf11 [label=n shape=ellipse]
\t}}
\tvalue_42 [label=42 shape=diamond]
\tattribute_n_on___config__Ainstance_{dir}maincf11 -> value_42
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf11 {{
\t\tattribute_requires_on___config__Ainstance_{dir}maincf11 [label=requires shape=ellipse]
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf11 {{
\t\tattribute_provides_on___config__Ainstance_{dir}maincf11 [label=provides shape=ellipse]
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf18 {{
\t\tlabel=A
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf18 {{
\t\tattribute_l_on___config__Ainstance_{dir}maincf18 [label=l shape=ellipse]
\t}}
\tassignable_y_{dir}maincf145 [label=y shape=ellipse]
\tassignable_x_{dir}maincf11 [label=x shape=ellipse]
\tassignable_x_{dir}maincf11 -> cluster_instance___config__Ainstance_{dir}maincf11
\tassignable_y_{dir}maincf145 -> assignable_x_{dir}maincf11
\tattribute_l_on___config__Ainstance_{dir}maincf18 -> assignable_y_{dir}maincf145 [label=".l"]
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf18 {{
\t\tattribute_m_on___config__Ainstance_{dir}maincf18 [label=m shape=ellipse]
\t}}
\tattribute_m_on___config__Ainstance_{dir}maincf18 -> assignable_x_{dir}maincf11 [label=".m"]
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf18 {{
\t\tattribute_n_on___config__Ainstance_{dir}maincf18 [label=n shape=ellipse]
\t}}
\tvalue_2 [label=2 shape=diamond]
\tattribute_n_on___config__Ainstance_{dir}maincf18 -> value_2
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf18 {{
\t\tattribute_requires_on___config__Ainstance_{dir}maincf18 [label=requires shape=ellipse]
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf18 {{
\t\tattribute_provides_on___config__Ainstance_{dir}maincf18 [label=provides shape=ellipse]
\t}}
\tassignable_u_{dir}maincf18 [label=u shape=ellipse]
\tassignable_u_{dir}maincf18 -> cluster_instance___config__Ainstance_{dir}maincf18
}}
        """,
    )


def test_dataflow_graphic_relation(graphic_asserter: GraphicAsserter) -> None:
    graphic_asserter(
        """
entity A:
end

entity B:
end

implement A using std::none
implement B using std::none

A.b [0:] -- B.a [0:]

a = A()
b = B()

a.b = b
        """,
        """
digraph {{
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf13 {{
\t\tlabel=A
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf13 {{
\t\tattribute_b_on___config__Ainstance_{dir}maincf13 [label=b shape=ellipse]
\t}}
\tassignable_b_{dir}maincf14 [label=b shape=ellipse]
\tsubgraph cluster_instance___config__Binstance_{dir}maincf14 {{
\t\tlabel=B
\t}}
\tsubgraph cluster_instance___config__Binstance_{dir}maincf14 {{
\t\tattribute_a_on___config__Binstance_{dir}maincf14 [label=a shape=ellipse]
\t}}
\tattribute_a_on___config__Binstance_{dir}maincf14 -> cluster_instance___config__Ainstance_{dir}maincf13
\tsubgraph cluster_instance___config__Binstance_{dir}maincf14 {{
\t\tattribute_requires_on___config__Binstance_{dir}maincf14 [label=requires shape=ellipse]
\t}}
\tsubgraph cluster_instance___config__Binstance_{dir}maincf14 {{
\t\tattribute_provides_on___config__Binstance_{dir}maincf14 [label=provides shape=ellipse]
\t}}
\tassignable_b_{dir}maincf14 -> cluster_instance___config__Binstance_{dir}maincf14
\tattribute_b_on___config__Ainstance_{dir}maincf13 -> assignable_b_{dir}maincf14
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf13 {{
\t\tattribute_requires_on___config__Ainstance_{dir}maincf13 [label=requires shape=ellipse]
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf13 {{
\t\tattribute_provides_on___config__Ainstance_{dir}maincf13 [label=provides shape=ellipse]
\t}}
\tassignable_a_{dir}maincf13 [label=a shape=ellipse]
\tassignable_a_{dir}maincf13 -> cluster_instance___config__Ainstance_{dir}maincf13
}}
        """,
    )


def test_dataflow_graphic_implementation(graphic_asserter: GraphicAsserter) -> None:
    graphic_asserter(
        """
entity A:
    number m
    number n
end

implement A using i

implementation i for A:
    self.m = self.n
end


A(n = 42)
A(n = 42)
        """,
        """
digraph {{
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf15 {{
\t\tlabel=A
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf15 {{
\t\tattribute_m_on___config__Ainstance_{dir}maincf15 [label=m shape=ellipse]
\t}}
\tassignable___self___{dir}maincf15 [label=__self__ shape=ellipse]
\tassignable___self___{dir}maincf15 -> cluster_instance___config__Ainstance_{dir}maincf15
\tattribute_m_on___config__Ainstance_{dir}maincf15 -> assignable___self___{dir}maincf15 [label=".n"]
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf15 {{
\t\tattribute_n_on___config__Ainstance_{dir}maincf15 [label=n shape=ellipse]
\t}}
\tvalue_42 [label=42 shape=diamond]
\tattribute_n_on___config__Ainstance_{dir}maincf15 -> value_42
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf15 {{
\t\tattribute_requires_on___config__Ainstance_{dir}maincf15 [label=requires shape=ellipse]
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf15 {{
\t\tattribute_provides_on___config__Ainstance_{dir}maincf15 [label=provides shape=ellipse]
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf14 {{
\t\tlabel=A
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf14 {{
\t\tattribute_m_on___config__Ainstance_{dir}maincf14 [label=m shape=ellipse]
\t}}
\tassignable___self___{dir}maincf14 [label=__self__ shape=ellipse]
\tassignable___self___{dir}maincf14 -> cluster_instance___config__Ainstance_{dir}maincf14
\tattribute_m_on___config__Ainstance_{dir}maincf14 -> assignable___self___{dir}maincf14 [label=".n"]
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf14 {{
\t\tattribute_n_on___config__Ainstance_{dir}maincf14 [label=n shape=ellipse]
\t}}
\tattribute_n_on___config__Ainstance_{dir}maincf14 -> value_42
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf14 {{
\t\tattribute_requires_on___config__Ainstance_{dir}maincf14 [label=requires shape=ellipse]
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf14 {{
\t\tattribute_provides_on___config__Ainstance_{dir}maincf14 [label=provides shape=ellipse]
\t}}
}}
        """,
    )


def test_dataflow_graphic_index(graphic_asserter: GraphicAsserter) -> None:
    graphic_asserter(
        """
entity A:
    number m
    number n
end

index A(n)

implement A using std::none


x = A(n = 42)
y = A(n = 42)

x.m = 0
        """,
        """
digraph {{
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf12 {{
\t\tlabel=A
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf13 {{
\t\tlabel=A
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf13 {{
\t\tattribute_m_on___config__Ainstance_{dir}maincf13 [label=m shape=ellipse]
\t}}
\tvalue_0 [label=0 shape=diamond]
\tattribute_m_on___config__Ainstance_{dir}maincf13 -> value_0
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf13 {{
\t\tattribute_n_on___config__Ainstance_{dir}maincf13 [label=n shape=ellipse]
\t}}
\tvalue_42 [label=42 shape=diamond]
\tattribute_n_on___config__Ainstance_{dir}maincf13 -> value_42
\tattribute_n_on___config__Ainstance_{dir}maincf13 -> value_42
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf13 {{
\t\tattribute_requires_on___config__Ainstance_{dir}maincf13 [label=requires shape=ellipse]
\t}}
\tsubgraph cluster_instance___config__Ainstance_{dir}maincf13 {{
\t\tattribute_provides_on___config__Ainstance_{dir}maincf13 [label=provides shape=ellipse]
\t}}
\tcluster_instance___config__Ainstance_{dir}maincf12 -> cluster_instance___config__Ainstance_{dir}maincf13 [label=index]
\tassignable_x_{dir}maincf12 [label=x shape=ellipse]
\tassignable_x_{dir}maincf12 -> cluster_instance___config__Ainstance_{dir}maincf12
\tassignable_y_{dir}maincf13 [label=y shape=ellipse]
\tassignable_y_{dir}maincf13 -> cluster_instance___config__Ainstance_{dir}maincf13
}}
        """,
    )



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import os

import pytest

from inmanta.config import Config
from inmanta.moduletool import ModuleTool
from moduletool.common import add_file, clone_repo


@pytest.mark.parametrize(
    "kwargs_update_method, mod2_should_be_updated, mod8_should_be_updated",
    [({}, True, True), ({"module": "mod2"}, True, False), ({"module": "mod8"}, False, True)],
)
def test_module_update_with_install_mode_master(
    tmpdir, modules_dir, modules_repo, kwargs_update_method, mod2_should_be_updated, mod8_should_be_updated
):
    # Make a copy of masterproject_multi_mod
    masterproject_multi_mod = tmpdir.join("masterproject_multi_mod")
    clone_repo(modules_repo, "masterproject_multi_mod", tmpdir)
    libs_folder = os.path.join(masterproject_multi_mod, "libs")
    os.mkdir(libs_folder)

    # Set masterproject_multi_mod as current project
    os.chdir(masterproject_multi_mod)
    os.curdir = masterproject_multi_mod
    Config.load_config()

    # Dependencies masterproject_multi_mod
    for mod in ["mod2", "mod8"]:
        # Clone mod in root tmpdir
        clone_repo(modules_repo, mod, tmpdir)

        # Clone mod from root of tmpdir into libs folder of masterproject_multi_mod
        clone_repo(tmpdir, mod, libs_folder)

        # Update module in root of tmpdir by adding an extra file
        file_name_extra_file = "test_file"
        path_mod = os.path.join(tmpdir, mod)
        add_file(path_mod, file_name_extra_file, "test", "Second commit")

        # Assert test_file not present in libs folder of masterproject_multi_mod
        path_extra_file = os.path.join(libs_folder, mod, file_name_extra_file)
        assert not os.path.exists(path_extra_file)

    # Update module(s) of masterproject_multi_mod
    ModuleTool().update(**kwargs_update_method)

    # Assert availability of test_file in masterproject_multi_mod
    extra_file_mod2 = os.path.join(libs_folder, "mod2", file_name_extra_file)
    assert os.path.exists(extra_file_mod2) == mod2_should_be_updated
    extra_file_mod8 = os.path.join(libs_folder, "mod8", file_name_extra_file)
    assert os.path.exists(extra_file_mod8) == mod8_should_be_updated



from builtins import super
from typing import List

from inmanta.server import SLICE_AGENT_MANAGER, SLICE_SERVER
from inmanta.server.extensions import ApplicationContext
from inmanta.server.protocol import Server, ServerSlice


class MyTestSlice(ServerSlice):
    def __init__(self):
        super(MyTestSlice, self).__init__("badplugin.badslice")

    def get_dependencies(self) -> List[str]:
        return [SLICE_SERVER, SLICE_AGENT_MANAGER]

    async def start(self) -> None:
        raise Exception("Too bad, this plugin is broken")


def setup(application: ApplicationContext) -> None:
    application.register_slice(MyTestSlice())
